/*
 * Helpers to convert between coordinates system.
 *
 * Dependencies:
 *     - Math.frag
 */

/**
 * This method performs an equirectangular projection, i.e.,
 * projects a 3D cartesian direction on a 2D cartesian coordinate system
 *
 * **Note**: The resulting coordinates are in the range [-0.5; 0.5].
 * Please use the `cartesianToEquirectangular` for coordinates in the
 * range [0.0; 1.0].
 *
 * @param dir The **normalized** coordinate to project
 * @returns The 2D projected direction
 */
vec2 cartesianToEquirectangularSymmetrical(vec3 dir) {
    float u = atan(dir.z, dir.x) * RECIPROCAL_PI2;
    float v = asin(dir.y) * RECIPROCAL_PI;
    return vec2(u, v);
}

/**
 * Overload of `cartesianToEquirectangularUnnormalized` with normalization
 */
vec2 cartesianToEquirectangular(vec3 dir) {
    return cartesianToEquirectangularSymmetrical(dir) + vec2(0.5);
}
