#ifdef EMULATE_PACKING

/* Input is usually mediump but converted to highp for enough precision up to
 * USHRT_MAX */
highp uint packUnorm2x16(highp vec2 v) {
    highp uvec2 expanded = uvec2(round(clamp(v, 0.0, 1.0)*65535.0));
    return (expanded.x & 0xffffu) | (expanded.y << 16u);
}

highp uint packSnorm2x16(highp vec2 v) {
    /* GLSL int -> uint cast preserves the bit pattern */
    highp uvec2 expanded = uvec2(ivec2(round(clamp(v, -1.0, 1.0)*32767.0)));
    return (expanded.x & 0xffffu) | (expanded.y << 16u);
}

mediump float unpackHalf(mediump uint val) {
    const highp float Magic = floatBitsToUint(113u << 23u);
    const mediump uint ShiftedExp = 0x7c00u;
    mediump uint e = val & ShiftedExp;

    highp uint r = val & 0x7FFFu;
    r <<= 13u;
    r += (127u - 15u) << 23u;

    /* Handle exponent special cases */
    if(e == ShiftedExp) { /* Inf/NaN? */
        r += (128u - 16u) << 23u; /* Extra exp adjust */
    } else if(e == 0u) { /* Zero/Denormal */
        r += 1u << 23u; /* Extra exp adjust */
        highp float f = uintBitsToFloat(r);
        f -= Magic;
        r = floatBitsToUint(f);
    }

    highp uint sign = (val & 0x8000u);
    r |= (sign << 16u);
    return uintBitsToFloat(r);
}

mediump vec2 unpackHalf2x16(highp uint val) {
    return vec2(unpackHalf(val & 0xFFFFu), unpackHalf(val >> 16u));
}

#endif

highp uvec2 packUnormVector(mediump vec4 v) {
    return uvec2(
        packUnorm2x16(v.xy),
        packUnorm2x16(v.zw)
    );
}

highp uvec2 packSnormVector(mediump vec4 v) {
    return uvec2(
        packSnorm2x16(v.xy),
        packSnorm2x16(v.zw)
    );
}
