/**
 * Types
 */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
import { nativeProperty, enumerable } from './decorators.js';
import { isImageLike, isNumber, isString } from './utils/object.js';
import { Emitter } from './utils/event.js';
import { Type, defaultPropertyCloner } from './property.js';
import { Resource, SceneResource } from './resources/resource.js';
import { Prefab } from './prefab.js';
import { createDestroyedProxy } from './utils/misc.js';
/**
 * Wonderland Engine API
 * @namespace WL
 */
/**
 * Default set of logging tags used by the API.
 */
export var LogTag;
(function (LogTag) {
    /** Initialization, component registration, etc... */
    LogTag[LogTag["Engine"] = 0] = "Engine";
    /** Scene loading */
    LogTag[LogTag["Scene"] = 1] = "Scene";
    /** Component init, update, etc... */
    LogTag[LogTag["Component"] = 2] = "Component";
})(LogTag || (LogTag = {}));
/**
 * Collider type enum for {@link CollisionComponent}.
 */
export var Collider;
(function (Collider) {
    /**
     * **Sphere Collider**:
     *
     * Simplest and most performant collision shape. If this type is set on a
     * {@link CollisionComponent}, only the first component of
     * {@link CollisionComponent#extents} will be used to determine the radius.
     */
    Collider[Collider["Sphere"] = 0] = "Sphere";
    /**
     * **Axis Aligned Bounding Box Collider**:
     *
     * Box that is always aligned to XYZ axis. It cannot be rotated but is more
     * efficient than {@link Collider.Box}.
     */
    Collider[Collider["AxisAlignedBox"] = 1] = "AxisAlignedBox";
    /**
     * **Aligned Bounding Box Collider**:
     *
     * Box that matches the object's rotation and translation correctly. This
     * is the least efficient collider and should only be chosen over
     * {@link Collider.Sphere} and {@link Collider.AxisAlignedBox} if really
     * necessary.
     */
    Collider[Collider["Box"] = 2] = "Box";
})(Collider || (Collider = {}));
/**
 * Alignment type enum for {@link TextComponent}.
 */
export var Alignment;
(function (Alignment) {
    /** Text start is at object origin */
    Alignment[Alignment["Left"] = 0] = "Left";
    /** Text center is at object origin */
    Alignment[Alignment["Center"] = 1] = "Center";
    /** Text end is at object origin */
    Alignment[Alignment["Right"] = 2] = "Right";
})(Alignment || (Alignment = {}));
/**
 * Vertical alignment type enum for {@link TextComponent}.
 */
export var VerticalAlignment;
(function (VerticalAlignment) {
    /** Text line is at object origin */
    VerticalAlignment[VerticalAlignment["Line"] = 0] = "Line";
    /** Text middle is at object origin */
    VerticalAlignment[VerticalAlignment["Middle"] = 1] = "Middle";
    /** Text top is at object origin */
    VerticalAlignment[VerticalAlignment["Top"] = 2] = "Top";
    /** Text bottom is at object origin */
    VerticalAlignment[VerticalAlignment["Bottom"] = 3] = "Bottom";
})(VerticalAlignment || (VerticalAlignment = {}));
/**
 * Justification type enum for {@link TextComponent}.
 *
 * @deprecated Please use {@link VerticalAlignment} instead.
 */
export const Justification = VerticalAlignment;
/**
 * Effect type enum for {@link TextComponent}.
 */
export var TextEffect;
(function (TextEffect) {
    /** Text is rendered normally */
    TextEffect[TextEffect["None"] = 0] = "None";
    /** Text is rendered with an outline */
    TextEffect[TextEffect["Outline"] = 1] = "Outline";
})(TextEffect || (TextEffect = {}));
/**
 * Wrap mode enum for {@link TextComponent}.
 *
 * @since 1.2.1
 */
export var TextWrapMode;
(function (TextWrapMode) {
    /** Text doesn't wrap automatically, only with explicit newline */
    TextWrapMode[TextWrapMode["None"] = 0] = "None";
    /** Text wraps at word boundaries */
    TextWrapMode[TextWrapMode["Soft"] = 1] = "Soft";
    /** Text wraps anywhere */
    TextWrapMode[TextWrapMode["Hard"] = 2] = "Hard";
    /** Text is cut off */
    TextWrapMode[TextWrapMode["Clip"] = 3] = "Clip";
})(TextWrapMode || (TextWrapMode = {}));
/**
 * Input type enum for {@link InputComponent}.
 */
export var InputType;
(function (InputType) {
    /** Head input */
    InputType[InputType["Head"] = 0] = "Head";
    /** Left eye input */
    InputType[InputType["EyeLeft"] = 1] = "EyeLeft";
    /** Right eye input */
    InputType[InputType["EyeRight"] = 2] = "EyeRight";
    /** Left controller input */
    InputType[InputType["ControllerLeft"] = 3] = "ControllerLeft";
    /** Right controller input */
    InputType[InputType["ControllerRight"] = 4] = "ControllerRight";
    /** Left ray input */
    InputType[InputType["RayLeft"] = 5] = "RayLeft";
    /** Right ray input */
    InputType[InputType["RayRight"] = 6] = "RayRight";
})(InputType || (InputType = {}));
/**
 * Light type enum for {@link LightComponent}.
 */
export var LightType;
(function (LightType) {
    /** Point light */
    LightType[LightType["Point"] = 0] = "Point";
    /** Spot light */
    LightType[LightType["Spot"] = 1] = "Spot";
    /** Sun light / Directional light */
    LightType[LightType["Sun"] = 2] = "Sun";
})(LightType || (LightType = {}));
/**
 * Animation state of {@link AnimationComponent}.
 */
export var AnimationState;
(function (AnimationState) {
    /** Animation is currently playing */
    AnimationState[AnimationState["Playing"] = 0] = "Playing";
    /** Animation is paused and will continue at current playback
     * time on {@link AnimationComponent#play} */
    AnimationState[AnimationState["Paused"] = 1] = "Paused";
    /** Animation is stopped */
    AnimationState[AnimationState["Stopped"] = 2] = "Stopped";
})(AnimationState || (AnimationState = {}));
/**
 * Rigid body force mode for {@link PhysXComponent#addForce} and {@link PhysXComponent#addTorque}.
 *
 * [PhysX API Reference](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxapi/files/structPxForceMode.html)
 */
export var ForceMode;
(function (ForceMode) {
    /** Apply as force */
    ForceMode[ForceMode["Force"] = 0] = "Force";
    /** Apply as impulse */
    ForceMode[ForceMode["Impulse"] = 1] = "Impulse";
    /** Apply as velocity change, mass dependent */
    ForceMode[ForceMode["VelocityChange"] = 2] = "VelocityChange";
    /** Apply as mass dependent force */
    ForceMode[ForceMode["Acceleration"] = 3] = "Acceleration";
})(ForceMode || (ForceMode = {}));
/**
 * Collision callback event type.
 */
export var CollisionEventType;
(function (CollisionEventType) {
    /** Touch/contact detected, collision */
    CollisionEventType[CollisionEventType["Touch"] = 0] = "Touch";
    /** Touch/contact lost, uncollide */
    CollisionEventType[CollisionEventType["TouchLost"] = 1] = "TouchLost";
    /** Touch/contact with trigger detected */
    CollisionEventType[CollisionEventType["TriggerTouch"] = 2] = "TriggerTouch";
    /** Touch/contact with trigger lost */
    CollisionEventType[CollisionEventType["TriggerTouchLost"] = 3] = "TriggerTouchLost";
})(CollisionEventType || (CollisionEventType = {}));
/**
 * Rigid body {@link PhysXComponent#shape}.
 *
 * [PhysX SDK Guide](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/Geometry.html#geometry-types).
 */
export var Shape;
(function (Shape) {
    /** No shape. */
    Shape[Shape["None"] = 0] = "None";
    /** Sphere shape. */
    Shape[Shape["Sphere"] = 1] = "Sphere";
    /** Capsule shape. */
    Shape[Shape["Capsule"] = 2] = "Capsule";
    /** Box shape. */
    Shape[Shape["Box"] = 3] = "Box";
    /** Plane shape. */
    Shape[Shape["Plane"] = 4] = "Plane";
    /** Convex mesh shape. */
    Shape[Shape["ConvexMesh"] = 5] = "ConvexMesh";
    /** Triangle mesh shape. */
    Shape[Shape["TriangleMesh"] = 6] = "TriangleMesh";
})(Shape || (Shape = {}));
/**
 * Mesh attribute enum.
 * @since 0.9.0
 */
export var MeshAttribute;
(function (MeshAttribute) {
    /** Position attribute, 3 floats */
    MeshAttribute[MeshAttribute["Position"] = 0] = "Position";
    /** Tangent attribute, 4 floats */
    MeshAttribute[MeshAttribute["Tangent"] = 1] = "Tangent";
    /** Normal attribute, 3 floats */
    MeshAttribute[MeshAttribute["Normal"] = 2] = "Normal";
    /** Texture coordinate attribute, 2 floats */
    MeshAttribute[MeshAttribute["TextureCoordinate"] = 3] = "TextureCoordinate";
    /** Color attribute, 4 floats, RGBA, range `0` to `1` */
    MeshAttribute[MeshAttribute["Color"] = 4] = "Color";
    /** Joint id attribute, 8 unsigned ints */
    MeshAttribute[MeshAttribute["JointId"] = 5] = "JointId";
    /** Joint weights attribute, 8 floats */
    MeshAttribute[MeshAttribute["JointWeight"] = 6] = "JointWeight";
})(MeshAttribute || (MeshAttribute = {}));
/** Proxy used to override prototypes of destroyed objects. */
export const DestroyedObjectInstance = createDestroyedProxy('object');
/** Proxy used to override prototypes of destroyed components. */
export const DestroyedComponentInstance = createDestroyedProxy('component');
/** Proxy used to override prototypes of destroyed prefabs. */
export const DestroyedPrefabInstance = createDestroyedProxy('prefab/scene');
/**
 * Check whether a given shape is a mesh or not.
 *
 * @param shape The shape to check.
 * @returns `true` if the shape is a mesh, `false` if it's a primitive.
 */
function isMeshShape(shape) {
    return shape === Shape.ConvexMesh || shape === Shape.TriangleMesh;
}
/**
 * Check whether an object is the {@link Component} class or not,
 *
 * @note This method if foolproof to developers inadvertently using
 * multiple Wonderland Engine API in a bundle.
 *
 * @param value The object to check.
 * @returns `true` if the object is a {@link Component} class, `false` otherwise.
 */
function isBaseComponentClass(value) {
    return (!!value &&
        value.hasOwnProperty('_isBaseComponent') &&
        value._isBaseComponent);
}
/**
 * Constants.
 */
/**
 * Default world up vector.
 */
const UP_VECTOR = [0, 1, 0];
const SQRT_3 = Math.sqrt(3);
/**
 * Provides access to a component instance of a specified component type.
 *
 * @example
 *
 * This is how you extend this class to create your own custom
 * component:
 *
 * ```js
 * import { Component, Type } from '@wonderlandengine/api';
 *
 * export class MyComponent extends Component {
 *     static TypeName = 'my-component';
 *     static Properties = {
 *         myBoolean: { type: Type.Boolean, default: false },
 *     };
 *     start() {}
 *     onActivate() {}
 *     onDeactivate() {}
 *     update(dt) {}
 * }
 * ```
 *
 * In a component, the scene can be accessed using `this.scene`:
 *
 * ```js
 * import { Component, Type } from '@wonderlandengine/api';
 *
 * export class MyComponent extends Component {
 *     static TypeName = 'my-component';
 *     start() {
 *         const obj = this.scene.addObject();
 *     }
 * }
 * ```
 */
export class Component {
    /**
     * Pack scene index and component id.
     *
     * @param scene Scene index.
     * @param id Component id.
     * @returns The packed id.
     *
     * @hidden
     */
    static _pack(scene, id) {
        return (scene << 22) | id;
    }
    /**
     * `true` for every class inheriting from this class.
     *
     * @note This is a workaround for `instanceof` to prevent issues
     * that could arise when an application ends up using multiple API versions.
     *
     * @hidden
     */
    static _isBaseComponent = true;
    /**
     * Fixed order of attributes in the `Properties` array.
     *
     * @note This is used for parameter deserialization and is filled during
     * component registration.
     *
     * @hidden
     */
    static _propertyOrder = [];
    /**
     * Unique identifier for this component class.
     *
     * This is used to register, add, and retrieve components of a given type.
     */
    static TypeName;
    /**
     * Properties of this component class.
     *
     * Properties are public attributes that can be configured via the
     * Wonderland Editor.
     *
     * Example:
     *
     * ```js
     * import { Component, Type } from '@wonderlandengine/api';
     * class MyComponent extends Component {
     *     static TypeName = 'my-component';
     *     static Properties = {
     *         myBoolean: { type: Type.Boolean, default: false },
     *         myFloat: { type: Type.Float, default: false },
     *         myTexture: { type: Type.Texture, default: null },
     *     };
     * }
     * ```
     *
     * Properties are automatically added to each component instance, and are
     * accessible like any JS attribute:
     *
     * ```js
     * // Creates a new component and set each properties value:
     * const myComponent = object.addComponent(MyComponent, {
     *     myBoolean: true,
     *     myFloat: 42.0,
     *     myTexture: null
     * });
     *
     * // You can also override the properties on the instance:
     * myComponent.myBoolean = false;
     * myComponent.myFloat = -42.0;
     * ```
     *
     * #### References
     *
     * Reference types (i.e., mesh, object, etc...) can also be listed as **required**:
     *
     * ```js
     * import {Component, Property} from '@wonderlandengine/api';
     *
     * class MyComponent extends Component {
     *     static Properties = {
     *         myObject: Property.object({required: true}),
     *         myAnimation: Property.animation({required: true}),
     *         myTexture: Property.texture({required: true}),
     *         myMesh: Property.mesh({required: true}),
     *     }
     * }
     * ```
     *
     * Please note that references are validated **once** before the call to {@link Component.start} only,
     * via the {@link Component.validateProperties} method.
     */
    static Properties;
    /**
     * When set to `true`, the child class inherits from the parent
     * properties, as shown in the following example:
     *
     * ```js
     * import {Component, Property} from '@wonderlandengine/api';
     *
     * class Parent extends Component {
     *     static TypeName = 'parent';
     *     static Properties = {parentName: Property.string('parent')}
     * }
     *
     * class Child extends Parent {
     *     static TypeName = 'child';
     *     static Properties = {name: Property.string('child')}
     *     static InheritProperties = true;
     *
     *     start() {
     *         // Works because `InheritProperties` is `true`.
     *         console.log(`${this.name} inherits from ${this.parentName}`);
     *     }
     * }
     * ```
     *
     * @note Properties defined in descendant classes will override properties
     * with the same name defined in ancestor classes.
     *
     * Defaults to `true`.
     */
    static InheritProperties;
    /**
     * Called when this component class is registered.
     *
     * @example
     *
     * This callback can be used to register dependencies of a component,
     * e.g., component classes that need to be registered in order to add
     * them at runtime with {@link Object3D.addComponent}, independent of whether
     * they are used in the editor.
     *
     * ```js
     * class Spawner extends Component {
     *     static TypeName = 'spawner';
     *
     *     static onRegister(engine) {
     *         engine.registerComponent(SpawnedComponent);
     *     }
     *
     *     // You can now use addComponent with SpawnedComponent
     * }
     * ```
     *
     * @example
     *
     * This callback can be used to register different implementations of a
     * component depending on client features or API versions.
     *
     * ```js
     * // Properties need to be the same for all implementations!
     * const SharedProperties = {};
     *
     * class Anchor extends Component {
     *     static TypeName = 'spawner';
     *     static Properties = SharedProperties;
     *
     *     static onRegister(engine) {
     *         if(navigator.xr === undefined) {
     *             /* WebXR unsupported, keep this dummy component *\/
     *             return;
     *         }
     *         /* WebXR supported! Override already registered dummy implementation
     *          * with one depending on hit-test API support *\/
     *         engine.registerComponent(window.HitTestSource === undefined ?
     *             AnchorWithoutHitTest : AnchorWithHitTest);
     *     }
     *
     *     // This one implements no functions
     * }
     * ```
     */
    static onRegister;
    /**
     * Allows to inherit properties directly inside the editor.
     *
     * @note Do not use directly, prefer using {@link inheritProperties}.
     *
     * @hidden
     */
    static _inheritProperties() {
        inheritProperties(this);
    }
    /** Manager index. @hidden */
    _manager;
    /** Packed id, containing the scene and the local id. @hidden */
    _id;
    /** Id relative to the scene component's manager. @hidden */
    _localId;
    /**
     * Object containing this object.
     *
     * **Note**: This is cached for faster retrieval.
     *
     * @hidden
     */
    _object;
    /** Scene instance. @hidden */
    _scene;
    /**
     * Create a new instance
     *
     * @param engine The engine instance.
     * @param manager Index of the manager.
     * @param id WASM component instance index.
     *
     * @hidden
     */
    constructor(scene, manager = -1, id = -1) {
        this._scene = scene;
        this._manager = manager;
        this._localId = id;
        this._id = Component._pack(scene._index, id);
        this._object = null;
    }
    /** Scene this component is part of. */
    get scene() {
        return this._scene;
    }
    /** Hosting engine instance. */
    get engine() {
        return this._scene.engine;
    }
    /** The name of this component's type */
    get type() {
        const ctor = this.constructor;
        return ctor.TypeName;
    }
    /** The object this component is attached to. */
    get object() {
        if (!this._object) {
            const objectId = this.engine.wasm._wl_component_get_object(this._manager, this._id);
            this._object = this._scene.wrap(objectId);
        }
        return this._object;
    }
    /**
     * Set whether this component is active.
     *
     * Activating/deactivating a component comes at a small cost of reordering
     * components in the respective component manager. This function therefore
     * is not a trivial assignment.
     *
     * Does nothing if the component is already activated/deactivated.
     *
     * @param active New active state.
     */
    set active(active) {
        this.engine.wasm._wl_component_setActive(this._manager, this._id, active);
    }
    /** `true` if the component is marked as active and its scene is active. */
    get active() {
        return this.markedActive && this._scene.isActive;
    }
    /**
     * `true` if the component is marked as active in the scene, `false` otherwise.
     *
     * @note At the opposite of {@link Component.active}, this accessor doesn't
     * take into account whether the scene is active or not.
     */
    get markedActive() {
        return this.engine.wasm._wl_component_isActive(this._manager, this._id) != 0;
    }
    /**
     * Copy all the properties from `src` into this instance.
     *
     * @note Only properties are copied. If a component needs to
     * copy extra data, it needs to override this method.
     *
     * #### Example
     *
     * ```js
     * class MyComponent extends Component {
     *     nonPropertyData = 'Hello World';
     *
     *     copy(src) {
     *         super.copy(src);
     *         this.nonPropertyData = src.nonPropertyData;
     *         return this;
     *     }
     * }
     * ```
     *
     * @note This method is called by {@link Object3D.clone}. Do not attempt to:
     *     - Create new component
     *     - Read references to other objects
     *
     * When cloning via {@link Object3D.clone}, this method will be called before
     * {@link Component.start}.
     *
     * @note JavaScript component properties aren't retargeted. Thus, references
     * inside the source object will not be retargeted to the destination object,
     * at the exception of the skin data on {@link MeshComponent} and {@link AnimationComponent}.
     *
     * @param src The source component to copy from.
     *
     * @returns Reference to self (for method chaining).
     */
    copy(src) {
        const ctor = this.constructor;
        const properties = ctor.Properties;
        if (!properties)
            return this;
        for (const name in properties) {
            const property = properties[name];
            const value = src[name];
            if (value === undefined)
                continue;
            const cloner = property.cloner ?? defaultPropertyCloner;
            this[name] = cloner.clone(property.type, value);
        }
        return this;
    }
    /**
     * Remove this component from its objects and destroy it.
     *
     * It is best practice to set the component to `null` after,
     * to ensure it does not get used later.
     *
     * ```js
     *    c.destroy();
     *    c = null;
     * ```
     * @since 0.9.0
     */
    destroy() {
        const manager = this._manager;
        if (manager < 0 || this._id < 0)
            return;
        /* This call will mark the component destroyed,
         * automatically calling `_triggerOnDestroy`.  */
        this.engine.wasm._wl_component_remove(manager, this._id);
    }
    /**
     * Checks equality by comparing ids and **not** the JavaScript reference.
     *
     * @deprecate Use JavaScript reference comparison instead:
     *
     * ```js
     * const componentA = obj.addComponent('mesh');
     * const componentB = obj.addComponent('mesh');
     * const componentC = componentB;
     * console.log(componentA === componentB); // false
     * console.log(componentA === componentA); // true
     * console.log(componentB === componentC); // true
     * ```
     */
    equals(otherComponent) {
        /** @todo(2.0.0): Remove this method. */
        if (!otherComponent)
            return false;
        return this._manager === otherComponent._manager && this._id === otherComponent._id;
    }
    /**
     * Reset the component properties to default.
     *
     * @note This is automatically called during the component instantiation.
     *
     * @returns Reference to self (for method chaining).
     */
    resetProperties() {
        const ctor = this.constructor;
        const properties = ctor.Properties;
        if (!properties)
            return this;
        for (const name in properties) {
            const property = properties[name];
            const cloner = property.cloner ?? defaultPropertyCloner;
            this[name] = cloner.clone(property.type, property.default);
        }
        return this;
    }
    /** @deprecated Use {@link Component.resetProperties} instead. */
    reset() {
        return this.resetProperties();
    }
    /**
     * Validate the properties on this instance.
     *
     * @throws If any of the required properties isn't initialized
     * on this instance.
     */
    validateProperties() {
        const ctor = this.constructor;
        if (!ctor.Properties)
            return;
        for (const name in ctor.Properties) {
            if (!ctor.Properties[name].required)
                continue;
            if (!this[name]) {
                throw new Error(`Property '${name}' is required but was not initialized`);
            }
        }
    }
    toString() {
        if (this.isDestroyed) {
            return 'Component(destroyed)';
        }
        return `Component('${this.type}', ${this._localId})`;
    }
    /**
     * `true` if the component is destroyed, `false` otherwise.
     *
     * If {@link WonderlandEngine.erasePrototypeOnDestroy} is `true`,
     * reading a custom property will not work:
     *
     * ```js
     * engine.erasePrototypeOnDestroy = true;
     *
     * const comp = obj.addComponent('mesh');
     * comp.customParam = 'Hello World!';
     *
     * console.log(comp.isDestroyed); // Prints `false`
     * comp.destroy();
     * console.log(comp.isDestroyed); // Prints `true`
     * console.log(comp.customParam); // Throws an error
     * ```
     *
     * @since 1.1.1
     */
    get isDestroyed() {
        return this._id < 0;
    }
    /** @hidden */
    _copy(src, offsetsPtr) {
        const wasm = this.engine.wasm;
        /** @todo: Support retargeting for `Object3D.clone` as well. */
        const offsets = wasm.HEAPU32;
        const offsetsStart = offsetsPtr >>> 2;
        const destScene = this._scene;
        const ctor = this.constructor;
        for (const name in ctor.Properties) {
            const value = src[name];
            if (value === null) {
                this[name] = null;
                continue;
            }
            const prop = ctor.Properties[name];
            const offset = offsets[offsetsStart + prop.type];
            let retargeted;
            switch (prop.type) {
                case Type.Object: {
                    const index = wasm._wl_object_index(value._id);
                    const id = wasm._wl_object_id(destScene._index, index + offset);
                    retargeted = destScene.wrap(id);
                    break;
                }
                case Type.Animation:
                    retargeted = destScene.animations.wrap(offset + value._index);
                    break;
                case Type.Skin:
                    retargeted = destScene.skins.wrap(offset + value._index);
                    break;
                default:
                    const cloner = prop.cloner ?? defaultPropertyCloner;
                    retargeted = cloner.clone(prop.type, value);
                    break;
            }
            this[name] = retargeted;
        }
        return this;
    }
    /**
     * Trigger the component {@link Component.init} method.
     *
     * @note Use this method instead of directly calling {@link Component.init},
     * because this method creates an handler for the {@link Component.start}.
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _triggerInit() {
        if (this.init) {
            try {
                this.init();
            }
            catch (e) {
                this.engine.log.error(LogTag.Component, `Exception during ${this.type} init() on object ${this.object.name}`);
                this.engine.log.error(LogTag.Component, e);
            }
        }
        /* Arm onActivate() with the initial start() call */
        const oldActivate = this.onActivate;
        this.onActivate = function () {
            this.onActivate = oldActivate;
            let failed = false;
            try {
                this.validateProperties();
            }
            catch (e) {
                this.engine.log.error(LogTag.Component, `Exception during ${this.type} validateProperties() on object ${this.object.name}`);
                this.engine.log.error(LogTag.Component, e);
                failed = true;
            }
            try {
                this.start?.();
            }
            catch (e) {
                this.engine.log.error(LogTag.Component, `Exception during ${this.type} start() on object ${this.object.name}`);
                this.engine.log.error(LogTag.Component, e);
                failed = true;
            }
            if (failed) {
                this.active = false;
                return;
            }
            if (!this.onActivate)
                return;
            try {
                this.onActivate();
            }
            catch (e) {
                this.engine.log.error(LogTag.Component, `Exception during ${this.type} onActivate() on object ${this.object.name}`);
                this.engine.log.error(LogTag.Component, e);
            }
        };
    }
    /**
     * Trigger the component {@link Component.update} method.
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _triggerUpdate(dt) {
        if (!this.update)
            return;
        try {
            this.update(dt);
        }
        catch (e) {
            this.engine.log.error(LogTag.Component, `Exception during ${this.type} update() on object ${this.object.name}`);
            this.engine.log.error(LogTag.Component, e);
            if (this.engine.wasm._deactivate_component_on_error) {
                this.active = false;
            }
        }
    }
    /**
     * Trigger the component {@link Component.onActivate} method.
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _triggerOnActivate() {
        if (!this.onActivate)
            return;
        try {
            this.onActivate();
        }
        catch (e) {
            this.engine.log.error(LogTag.Component, `Exception during ${this.type} onActivate() on object ${this.object.name}`);
            this.engine.log.error(LogTag.Component, e);
        }
    }
    /**
     * Trigger the component {@link Component.onDeactivate} method.
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _triggerOnDeactivate() {
        if (!this.onDeactivate)
            return;
        try {
            this.onDeactivate();
        }
        catch (e) {
            this.engine.log.error(LogTag.Component, `Exception during ${this.type} onDeactivate() on object ${this.object.name}`);
            this.engine.log.error(LogTag.Component, e);
        }
    }
    /**
     * Trigger the component {@link Component.onDestroy} method.
     *
     * @note This api is meant to be used internally.
     *
     * @hidden
     */
    _triggerOnDestroy() {
        try {
            if (this.onDestroy)
                this.onDestroy();
        }
        catch (e) {
            this.engine.log.error(LogTag.Component, `Exception during ${this.type} onDestroy() on object ${this.object.name}`);
            this.engine.log.error(LogTag.Component, e);
        }
        this._scene._components.destroy(this);
    }
}
/**
 * Components must be registered before loading / appending a scene.
 *
 * It's possible to end up with a broken component in the following cases:
 *
 * - Component wasn't registered when the scene was loaded
 * - Component instantiation failed
 *
 * This dummy component is thus used as a placeholder by the engine.
 */
export class BrokenComponent extends Component {
    static TypeName = '__broken-component__';
}
/**
 * Merge the ascendant properties of class
 *
 * This method walks the prototype chain, and merges
 * all the properties found in parent components.
 *
 * Example:
 *
 * ```js
 * import {Property, inheritProperties} from '@wonderlandengine/api';
 *
 * class Parent {
 *     static Properties = { parentProp: Property.string('parent') };
 * }
 *
 * class Child extends Parent {
 *     static Properties = { childProp: Property.string('child') };
 * }
 * inheritProperties(Child);
 * ```
 *
 * @param target The class in which properties should be merged
 *
 * @hidden
 */
export function inheritProperties(target) {
    if (!target.TypeName)
        return;
    const chain = [];
    let curr = target;
    while (curr && !isBaseComponentClass(curr)) {
        const comp = curr;
        /* Stop at the first class that doesn't require properties merging */
        const needsMerge = comp.hasOwnProperty('InheritProperties')
            ? comp.InheritProperties
            : true;
        if (!needsMerge)
            break;
        if (comp.TypeName && comp.hasOwnProperty('Properties')) {
            chain.push(comp);
        }
        curr = Object.getPrototypeOf(curr);
    }
    /* No prototype merge is needed. */
    if (!chain.length || (chain.length === 1 && chain[0] === target)) {
        return;
    }
    const merged = {};
    for (let i = chain.length - 1; i >= 0; --i) {
        Object.assign(merged, chain[i].Properties);
    }
    target.Properties = merged;
}
/**
 * Native collision component.
 *
 * Provides access to a native collision component instance.
 */
export class CollisionComponent extends Component {
    /** @override */
    static TypeName = 'collision';
    getExtents(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_collision_component_get_extents(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        return out;
    }
    /** Collision component collider */
    get collider() {
        return this.engine.wasm._wl_collision_component_get_collider(this._id);
    }
    /**
     * Set collision component collider.
     *
     * @param collider Collider of the collision component.
     */
    set collider(collider) {
        this.engine.wasm._wl_collision_component_set_collider(this._id, collider);
    }
    /**
     * Equivalent to {@link CollisionComponent.getExtents}.
     *
     * @note Prefer to use {@link CollisionComponent.getExtents} for performance.
     */
    get extents() {
        /** @todo: Break at 2.0.0. Do not allow modifying memory in-place. */
        const wasm = this.engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_collision_component_get_extents(this._id), 3);
    }
    /**
     * Set collision component extents.
     *
     * If {@link collider} returns {@link Collider.Sphere}, only the first
     * component of the passed vector is used.
     *
     * Example:
     *
     * ```js
     * // Spans 1 unit on the x-axis, 2 on the y-axis, 3 on the z-axis.
     * collision.extent = [1, 2, 3];
     * ```
     *
     * @param extents Extents of the collision component, expects a
     *      3 component array.
     */
    set extents(extents) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_collision_component_get_extents(this._id) / 4; /* Align F32 */
        wasm.HEAPF32[ptr] = extents[0];
        wasm.HEAPF32[ptr + 1] = extents[1];
        wasm.HEAPF32[ptr + 2] = extents[2];
    }
    /**
     * Get collision component radius.
     *
     * @note If {@link collider} is not {@link Collider.Sphere}, the returned value
     * corresponds to the radius of a sphere enclosing the shape.
     *
     * Example:
     *
     * ```js
     * sphere.radius = 3.0;
     * console.log(sphere.radius); // 3.0
     *
     * box.extents = [2.0, 2.0, 2.0];
     * console.log(box.radius); // 1.732...
     * ```
     *
     */
    get radius() {
        const wasm = this.engine.wasm;
        if (this.collider === Collider.Sphere)
            return wasm.HEAPF32[wasm._wl_collision_component_get_extents(this._id) >> 2];
        const extents = new Float32Array(wasm.HEAPF32.buffer, wasm._wl_collision_component_get_extents(this._id), 3);
        const x2 = extents[0] * extents[0];
        const y2 = extents[1] * extents[1];
        const z2 = extents[2] * extents[2];
        return Math.sqrt(x2 + y2 + z2) / 2;
    }
    /**
     * Set collision component radius.
     *
     * @param radius Radius of the collision component
     *
     * @note If {@link collider} is not {@link Collider.Sphere},
     * the extents are set to form a square that fits a sphere with the provided radius.
     *
     * Example:
     *
     * ```js
     * aabbCollision.radius = 2.0; // AABB fits a sphere of radius 2.0
     * boxCollision.radius = 3.0; // Box now fits a sphere of radius 3.0, keeping orientation
     * ```
     *
     */
    set radius(radius) {
        const length = this.collider === Collider.Sphere ? radius : (2 * radius) / SQRT_3;
        this.extents.set([length, length, length]);
    }
    /**
     * Collision component group.
     *
     * The groups is a bitmask that is compared to other components in {@link CollisionComponent#queryOverlaps}
     * or the group in {@link Scene#rayCast}.
     *
     * Colliders that have no common groups will not overlap with each other. If a collider
     * has none of the groups set for {@link Scene#rayCast}, the ray will not hit it.
     *
     * Each bit represents belonging to a group, see example.
     *
     * ```js
     *    // c belongs to group 2
     *    c.group = (1 << 2);
     *
     *    // c belongs to group 0
     *    c.group = (1 << 0);
     *
     *    // c belongs to group 0 *and* 2
     *    c.group = (1 << 0) | (1 << 2);
     *
     *    (c.group & (1 << 2)) != 0; // true
     *    (c.group & (1 << 7)) != 0; // false
     * ```
     */
    get group() {
        return this.engine.wasm._wl_collision_component_get_group(this._id);
    }
    /**
     * Set collision component group.
     *
     * @param group Group mask of the collision component.
     */
    set group(group) {
        this.engine.wasm._wl_collision_component_set_group(this._id, group);
    }
    /**
     * Query overlapping objects.
     *
     * Usage:
     *
     * ```js
     * const collision = object.getComponent('collision');
     * const overlaps = collision.queryOverlaps();
     * for(const otherCollision of overlaps) {
     *     const otherObject = otherCollision.object;
     *     console.log(`Collision with object ${otherObject.objectId}`);
     * }
     * ```
     *
     * @returns Collision components overlapping this collider.
     */
    queryOverlaps() {
        const count = this.engine.wasm._wl_collision_component_query_overlaps(this._id, this.engine.wasm._tempMem, this.engine.wasm._tempMemSize >> 1);
        const overlaps = new Array(count);
        for (let i = 0; i < count; ++i) {
            const id = this.engine.wasm._tempMemUint16[i];
            overlaps[i] = this._scene._components.wrapCollision(id);
        }
        return overlaps;
    }
}
__decorate([
    nativeProperty()
], CollisionComponent.prototype, "collider", null);
__decorate([
    nativeProperty()
], CollisionComponent.prototype, "extents", null);
__decorate([
    nativeProperty()
], CollisionComponent.prototype, "group", null);
/**
 * Native text component
 *
 * Provides access to a native text component instance
 */
export class TextComponent extends Component {
    /** @override */
    static TypeName = 'text';
    /** Text component alignment. */
    get alignment() {
        return this.engine.wasm._wl_text_component_get_horizontal_alignment(this._id);
    }
    /**
     * Set text component alignment.
     *
     * @param alignment Alignment for the text component.
     */
    set alignment(alignment) {
        this.engine.wasm._wl_text_component_set_horizontal_alignment(this._id, alignment);
    }
    /**
     * Text component vertical alignment.
     * @since 1.2.0
     */
    get verticalAlignment() {
        return this.engine.wasm._wl_text_component_get_vertical_alignment(this._id);
    }
    /**
     * Set text component vertical alignment.
     *
     * @param verticalAlignment Vertical for the text component.
     * @since 1.2.0
     */
    set verticalAlignment(verticalAlignment) {
        this.engine.wasm._wl_text_component_set_vertical_alignment(this._id, verticalAlignment);
    }
    /**
     * Text component justification.
     *
     * @deprecated Please use {@link TextComponent.verticalAlignment} instead.
     */
    get justification() {
        return this.verticalAlignment;
    }
    /**
     * Set text component justification.
     *
     * @param justification Justification for the text component.
     *
     * @deprecated Please use {@link TextComponent.verticalAlignment} instead.
     */
    set justification(justification) {
        this.verticalAlignment = justification;
    }
    /** Text component character spacing. */
    get characterSpacing() {
        return this.engine.wasm._wl_text_component_get_character_spacing(this._id);
    }
    /**
     * Set text component character spacing.
     *
     * @param spacing Character spacing for the text component.
     */
    set characterSpacing(spacing) {
        this.engine.wasm._wl_text_component_set_character_spacing(this._id, spacing);
    }
    /** Text component line spacing. */
    get lineSpacing() {
        return this.engine.wasm._wl_text_component_get_line_spacing(this._id);
    }
    /**
     * Set text component line spacing
     *
     * @param spacing Line spacing for the text component
     */
    set lineSpacing(spacing) {
        this.engine.wasm._wl_text_component_set_line_spacing(this._id, spacing);
    }
    /** Text component effect. */
    get effect() {
        return this.engine.wasm._wl_text_component_get_effect(this._id);
    }
    /**
     * Set text component effect
     *
     * @param effect Effect for the text component
     */
    set effect(effect) {
        this.engine.wasm._wl_text_component_set_effect(this._id, effect);
    }
    /**
     * Text component line wrap mode.
     * @since 1.2.1
     */
    get wrapMode() {
        return this.engine.wasm._wl_text_component_get_wrapMode(this._id);
    }
    /**
     * Set text component line wrap mode.
     *
     * @param wrapMode Line wrap mode for the text component.
     * @since 1.2.1
     */
    set wrapMode(wrapMode) {
        this.engine.wasm._wl_text_component_set_wrapMode(this._id, wrapMode);
    }
    /**
     * Text component line wrap width.
     * @since 1.2.1
     */
    get wrapWidth() {
        return this.engine.wasm._wl_text_component_get_wrapWidth(this._id);
    }
    /**
     * Set text component line wrap width.
     *
     * Only takes effect when {@link wrapMode} is something other than
     * {@link TextWrapMode.None}.
     *
     * @param width Line wrap width for the text component.
     * @since 1.2.1
     */
    set wrapWidth(width) {
        this.engine.wasm._wl_text_component_set_wrapWidth(this._id, width);
    }
    /** Text component text. */
    get text() {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_text_component_get_text(this._id);
        return wasm.UTF8ToString(ptr);
    }
    /**
     * Set text component text.
     *
     * @param text Text of the text component.
     */
    set text(text) {
        const wasm = this.engine.wasm;
        wasm._wl_text_component_set_text(this._id, wasm.tempUTF8(text.toString()));
    }
    /**
     * Set material to render the text with.
     *
     * @param material New material.
     */
    set material(material) {
        const matIndex = material ? material._id : 0;
        this.engine.wasm._wl_text_component_set_material(this._id, matIndex);
    }
    /** Material used to render the text. */
    get material() {
        const index = this.engine.wasm._wl_text_component_get_material(this._id);
        return this.engine.materials.wrap(index);
    }
    /** @overload */
    getBoundingBoxForText(text, out = new Float32Array(4)) {
        const wasm = this.engine.wasm;
        /* Offset by 4 floats, output is written to _tempMem */
        const textPtr = wasm.tempUTF8(text, 4 * 4);
        this.engine.wasm._wl_text_component_get_boundingBox(this._id, textPtr, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        out[3] = wasm._tempMemFloat[3];
        return out;
    }
    /** @overload */
    getBoundingBox(out = new Float32Array(4)) {
        const wasm = this.engine.wasm;
        this.engine.wasm._wl_text_component_get_boundingBox(this._id, 0, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        out[3] = wasm._tempMemFloat[3];
        return out;
    }
}
__decorate([
    nativeProperty()
], TextComponent.prototype, "alignment", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "verticalAlignment", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "justification", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "characterSpacing", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "lineSpacing", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "effect", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "wrapMode", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "wrapWidth", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "text", null);
__decorate([
    nativeProperty()
], TextComponent.prototype, "material", null);
/**
 * Native view component.
 *
 * Provides access to a native view component instance.
 */
export class ViewComponent extends Component {
    /** @override */
    static TypeName = 'view';
    getProjectionMatrix(out = new Float32Array(16)) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_view_component_get_projection_matrix(this._id) / 4; /* Align F32 */
        for (let i = 0; i < 16; ++i) {
            out[i] = wasm.HEAPF32[ptr + i];
        }
        return out;
    }
    /**
     * Equivalent to {@link ViewComponent.getProjectionMatrix}.
     *
     * @note Prefer to use {@link ViewComponent.getProjectionMatrix} for performance.
     */
    get projectionMatrix() {
        /** @todo: Break at 2.0.0. Do not allow modifying memory in-place. */
        const wasm = this.engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_view_component_get_projection_matrix(this._id), 16);
    }
    /** ViewComponent near clipping plane value. */
    get near() {
        return this.engine.wasm._wl_view_component_get_near(this._id);
    }
    /**
     * Set near clipping plane distance for the view.
     *
     * If an XR session is active, the change will apply in the
     * following frame, otherwise the change is immediate.
     *
     * @param near Near depth value.
     */
    set near(near) {
        this.engine.wasm._wl_view_component_set_near(this._id, near);
    }
    /** Far clipping plane value. */
    get far() {
        return this.engine.wasm._wl_view_component_get_far(this._id);
    }
    /**
     * Set far clipping plane distance for the view.
     *
     * If an XR session is active, the change will apply in the
     * following frame, otherwise the change is immediate.
     *
     * @param far Near depth value.
     */
    set far(far) {
        this.engine.wasm._wl_view_component_set_far(this._id, far);
    }
    /**
     * Get the horizontal field of view for the view, **in degrees**.
     *
     * If an XR session is active, this returns the field of view reported by
     * the device, regardless of the fov that was set.
     */
    get fov() {
        return this.engine.wasm._wl_view_component_get_fov(this._id);
    }
    /**
     * Set the horizontal field of view for the view, **in degrees**.
     *
     * If an XR session is active, the field of view reported by the device is
     * used and this value is ignored. After the XR session ends, the new value
     * is applied.
     *
     * @param fov Horizontal field of view, **in degrees**.
     */
    set fov(fov) {
        this.engine.wasm._wl_view_component_set_fov(this._id, fov);
    }
}
__decorate([
    enumerable()
], ViewComponent.prototype, "projectionMatrix", null);
__decorate([
    nativeProperty()
], ViewComponent.prototype, "near", null);
__decorate([
    nativeProperty()
], ViewComponent.prototype, "far", null);
__decorate([
    nativeProperty()
], ViewComponent.prototype, "fov", null);
/**
 * Native input component.
 *
 * Provides access to a native input component instance.
 */
export class InputComponent extends Component {
    /** @override */
    static TypeName = 'input';
    /** Input component type */
    get inputType() {
        return this.engine.wasm._wl_input_component_get_type(this._id);
    }
    /**
     * Set input component type.
     *
     * @params New input component type.
     */
    set inputType(type) {
        this.engine.wasm._wl_input_component_set_type(this._id, type);
    }
    /**
     * WebXR Device API input source associated with this input component,
     * if type {@link InputType.ControllerLeft} or {@link InputType.ControllerRight}.
     */
    get xrInputSource() {
        const xr = this.engine.xr;
        if (!xr)
            return null;
        for (let inputSource of xr.session.inputSources) {
            if (inputSource.handedness == this.handedness) {
                return inputSource;
            }
        }
        return null;
    }
    /**
     * 'left', 'right' or `null` depending on the {@link InputComponent#inputType}.
     */
    get handedness() {
        const inputType = this.inputType;
        if (inputType == InputType.ControllerRight ||
            inputType == InputType.RayRight ||
            inputType == InputType.EyeRight)
            return 'right';
        if (inputType == InputType.ControllerLeft ||
            inputType == InputType.RayLeft ||
            inputType == InputType.EyeLeft)
            return 'left';
        return null;
    }
}
__decorate([
    nativeProperty()
], InputComponent.prototype, "inputType", null);
__decorate([
    enumerable()
], InputComponent.prototype, "xrInputSource", null);
__decorate([
    enumerable()
], InputComponent.prototype, "handedness", null);
/**
 * Native light component.
 *
 * Provides access to a native light component instance.
 */
export class LightComponent extends Component {
    /** @override */
    static TypeName = 'light';
    getColor(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_light_component_get_color(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        return out;
    }
    /**
     * Set light color.
     *
     * @param c New color array/vector, expected to have at least 3 elements.
     * @since 1.0.0
     */
    setColor(c) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_light_component_get_color(this._id) / 4; /* Align F32 */
        wasm.HEAPF32[ptr] = c[0];
        wasm.HEAPF32[ptr + 1] = c[1];
        wasm.HEAPF32[ptr + 2] = c[2];
    }
    /**
     * View on the light color.
     *
     * @note Prefer to use {@link getColor} in performance-critical code.
     */
    get color() {
        const wasm = this.engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_light_component_get_color(this._id), 3);
    }
    /**
     * Set light color.
     *
     * @param c Color of the light component.
     *
     * @note Prefer to use {@link setColor} in performance-critical code.
     */
    set color(c) {
        this.color.set(c);
    }
    /** Light type. */
    get lightType() {
        return this.engine.wasm._wl_light_component_get_type(this._id);
    }
    /**
     * Set light type.
     *
     * @param lightType Type of the light component.
     */
    set lightType(t) {
        this.engine.wasm._wl_light_component_set_type(this._id, t);
    }
    /**
     * Light intensity.
     * @since 1.0.0
     */
    get intensity() {
        return this.engine.wasm._wl_light_component_get_intensity(this._id);
    }
    /**
     * Set light intensity.
     *
     * @param intensity Intensity of the light component.
     * @since 1.0.0
     */
    set intensity(intensity) {
        this.engine.wasm._wl_light_component_set_intensity(this._id, intensity);
    }
    /**
     * Outer angle for spot lights, in degrees.
     * @since 1.0.0
     */
    get outerAngle() {
        return this.engine.wasm._wl_light_component_get_outerAngle(this._id);
    }
    /**
     * Set outer angle for spot lights.
     *
     * @param angle Outer angle, in degrees.
     * @since 1.0.0
     */
    set outerAngle(angle) {
        this.engine.wasm._wl_light_component_set_outerAngle(this._id, angle);
    }
    /**
     * Inner angle for spot lights, in degrees.
     * @since 1.0.0
     */
    get innerAngle() {
        return this.engine.wasm._wl_light_component_get_innerAngle(this._id);
    }
    /**
     * Set inner angle for spot lights.
     *
     * @param angle Inner angle, in degrees.
     * @since 1.0.0
     */
    set innerAngle(angle) {
        this.engine.wasm._wl_light_component_set_innerAngle(this._id, angle);
    }
    /**
     * Whether the light casts shadows.
     * @since 1.0.0
     */
    get shadows() {
        return !!this.engine.wasm._wl_light_component_get_shadows(this._id);
    }
    /**
     * Set whether the light casts shadows.
     *
     * @param b Whether the light casts shadows.
     * @since 1.0.0
     */
    set shadows(b) {
        this.engine.wasm._wl_light_component_set_shadows(this._id, b);
    }
    /**
     * Range for shadows.
     * @since 1.0.0
     */
    get shadowRange() {
        return this.engine.wasm._wl_light_component_get_shadowRange(this._id);
    }
    /**
     * Set range for shadows.
     *
     * @param range Range for shadows.
     * @since 1.0.0
     */
    set shadowRange(range) {
        this.engine.wasm._wl_light_component_set_shadowRange(this._id, range);
    }
    /**
     * Bias value for shadows.
     * @since 1.0.0
     */
    get shadowBias() {
        return this.engine.wasm._wl_light_component_get_shadowBias(this._id);
    }
    /**
     * Set bias value for shadows.
     *
     * @param bias Bias for shadows.
     * @since 1.0.0
     */
    set shadowBias(bias) {
        this.engine.wasm._wl_light_component_set_shadowBias(this._id, bias);
    }
    /**
     * Normal bias value for shadows.
     * @since 1.0.0
     */
    get shadowNormalBias() {
        return this.engine.wasm._wl_light_component_get_shadowNormalBias(this._id);
    }
    /**
     * Set normal bias value for shadows.
     *
     * @param bias Normal bias for shadows.
     * @since 1.0.0
     */
    set shadowNormalBias(bias) {
        this.engine.wasm._wl_light_component_set_shadowNormalBias(this._id, bias);
    }
    /**
     * Texel size for shadows.
     * @since 1.0.0
     */
    get shadowTexelSize() {
        return this.engine.wasm._wl_light_component_get_shadowTexelSize(this._id);
    }
    /**
     * Set texel size for shadows.
     *
     * @param size Texel size for shadows.
     * @since 1.0.0
     */
    set shadowTexelSize(size) {
        this.engine.wasm._wl_light_component_set_shadowTexelSize(this._id, size);
    }
    /**
     * Cascade count for {@link LightType.Sun} shadows.
     * @since 1.0.0
     */
    get cascadeCount() {
        return this.engine.wasm._wl_light_component_get_cascadeCount(this._id);
    }
    /**
     * Set cascade count for {@link LightType.Sun} shadows.
     *
     * @param count Cascade count.
     * @since 1.0.0
     */
    set cascadeCount(count) {
        this.engine.wasm._wl_light_component_set_cascadeCount(this._id, count);
    }
}
__decorate([
    nativeProperty()
], LightComponent.prototype, "color", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "lightType", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "intensity", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "outerAngle", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "innerAngle", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "shadows", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "shadowRange", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "shadowBias", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "shadowNormalBias", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "shadowTexelSize", null);
__decorate([
    nativeProperty()
], LightComponent.prototype, "cascadeCount", null);
/**
 * Native animation component.
 *
 * Provides access to a native animation component instance.
 */
export class AnimationComponent extends Component {
    /** @override */
    static TypeName = 'animation';
    /**
     * Emitter for animation events triggered on this component.
     *
     * The first argument is the name of the event.
     */
    onEvent = new Emitter();
    /**
     * Set animation to play.
     *
     * Make sure to {@link Animation#retarget} the animation to affect the
     * right objects.
     *
     * @param anim Animation to play.
     */
    set animation(anim) {
        this.scene.assertOrigin(anim);
        this.engine.wasm._wl_animation_component_set_animation(this._id, anim ? anim._id : 0);
    }
    /** Animation set for this component */
    get animation() {
        const index = this.engine.wasm._wl_animation_component_get_animation(this._id);
        return this._scene.animations.wrap(index);
    }
    /**
     * Set play count. Set to `0` to loop indefinitely.
     *
     * @param playCount Number of times to repeat the animation.
     */
    set playCount(playCount) {
        this.engine.wasm._wl_animation_component_set_playCount(this._id, playCount);
    }
    /** Number of times the animation is played. */
    get playCount() {
        return this.engine.wasm._wl_animation_component_get_playCount(this._id);
    }
    /**
     * Set speed. Set to negative values to run the animation backwards.
     *
     * Setting speed has an immediate effect for the current frame's update
     * and will continue with the speed from the current point in the animation.
     *
     * @param speed New speed at which to play the animation.
     * @since 0.8.10
     */
    set speed(speed) {
        this.engine.wasm._wl_animation_component_set_speed(this._id, speed);
    }
    /**
     * Speed factor at which the animation is played.
     *
     * @since 0.8.10
     */
    get speed() {
        return this.engine.wasm._wl_animation_component_get_speed(this._id);
    }
    /** Current playing state of the animation */
    get state() {
        return this.engine.wasm._wl_animation_component_state(this._id);
    }
    /**
     * Play animation.
     *
     * If the animation is currently paused, resumes from that position. If the
     * animation is already playing, does nothing.
     *
     * To restart the animation, {@link AnimationComponent#stop} it first.
     */
    play() {
        this.engine.wasm._wl_animation_component_play(this._id);
    }
    /** Stop animation. */
    stop() {
        this.engine.wasm._wl_animation_component_stop(this._id);
    }
    /** Pause animation. */
    pause() {
        this.engine.wasm._wl_animation_component_pause(this._id);
    }
    /**
     * Get the value of a float parameter in the attached graph.
     * Throws if the parameter is missing.
     *
     * @param name Name of the parameter.
     * @since 1.2.0
     */
    getFloatParameter(name) {
        const wasm = this.engine.wasm;
        const index = wasm._wl_animation_component_getGraphParamIndex(this._id, wasm.tempUTF8(name));
        if (index === -1) {
            throw Error(`Missing parameter '${name}'`);
        }
        wasm._wl_animation_component_getGraphParamValue(this._id, index, wasm._tempMem);
        return wasm._tempMemFloat[0];
    }
    /**
     * Set the value of a float parameter in the attached graph
     * Throws if the parameter is missing.
     *
     * @param name Name of the parameter.
     * @param value Float value to set.
     * @returns 1 if the parameter was successfully set, 0 on fail.
     * @since 1.2.0
     */
    setFloatParameter(name, value) {
        const wasm = this.engine.wasm;
        const index = wasm._wl_animation_component_getGraphParamIndex(this._id, wasm.tempUTF8(name));
        if (index === -1) {
            throw Error(`Missing parameter '${name}'`);
        }
        wasm._tempMemFloat[0] = value;
        wasm._wl_animation_component_setGraphParamValue(this._id, index, wasm._tempMem);
    }
}
__decorate([
    nativeProperty()
], AnimationComponent.prototype, "animation", null);
__decorate([
    nativeProperty()
], AnimationComponent.prototype, "playCount", null);
__decorate([
    nativeProperty()
], AnimationComponent.prototype, "speed", null);
__decorate([
    enumerable()
], AnimationComponent.prototype, "state", null);
/**
 * Native mesh component.
 *
 * Provides access to a native mesh component instance.
 */
export class MeshComponent extends Component {
    /** @override */
    static TypeName = 'mesh';
    /**
     * Set material to render the mesh with.
     *
     * @param material Material to render the mesh with.
     */
    set material(material) {
        this.engine.wasm._wl_mesh_component_set_material(this._id, material ? material._id : 0);
    }
    /** Material used to render the mesh. */
    get material() {
        const index = this.engine.wasm._wl_mesh_component_get_material(this._id);
        return this.engine.materials.wrap(index);
    }
    /** Mesh rendered by this component. */
    get mesh() {
        const index = this.engine.wasm._wl_mesh_component_get_mesh(this._id);
        return this.engine.meshes.wrap(index);
    }
    /**
     * Set mesh to rendered with this component.
     *
     * @param mesh Mesh rendered by this component.
     */
    set mesh(mesh) {
        this.engine.wasm._wl_mesh_component_set_mesh(this._id, mesh?._id ?? 0);
    }
    /** Skin for this mesh component. */
    get skin() {
        const index = this.engine.wasm._wl_mesh_component_get_skin(this._id);
        return this._scene.skins.wrap(index);
    }
    /**
     * Set skin to transform this mesh component.
     *
     * @param skin Skin to use for rendering skinned meshes.
     */
    set skin(skin) {
        this.scene.assertOrigin(skin);
        this.engine.wasm._wl_mesh_component_set_skin(this._id, skin ? skin._id : 0);
    }
    /**
     * Morph targets for this mesh component.
     *
     * @since 1.2.0
     */
    get morphTargets() {
        const index = this.engine.wasm._wl_mesh_component_get_morph_targets(this._id);
        return this.engine.morphTargets.wrap(index);
    }
    /**
     * Set morph targets to transform this mesh component.
     *
     * @param morphTargets Morph targets to use for rendering.
     *
     * @since 1.2.0
     */
    set morphTargets(morphTargets) {
        this.engine.wasm._wl_mesh_component_set_morph_targets(this._id, morphTargets?._id ?? 0);
    }
    /**
     * Equivalent to {@link getMorphTargetWeights}.
     *
     * @note Prefer to use {@link getMorphTargetWeights} for performance.
     *
     * @since 1.2.0
     */
    get morphTargetWeights() {
        return this.getMorphTargetWeights();
    }
    /**
     * Set the morph target weights to transform this mesh component.
     *
     * @param weights New weights.
     *
     * @since 1.2.0
     */
    set morphTargetWeights(weights) {
        this.setMorphTargetWeights(weights);
    }
    getMorphTargetWeights(out) {
        const wasm = this.engine.wasm;
        const count = wasm._wl_mesh_component_get_morph_target_weights(this._id, wasm._tempMem);
        if (!out) {
            out = new Float32Array(count);
        }
        for (let i = 0; i < count; ++i) {
            out[i] = wasm._tempMemFloat[i];
        }
        return out;
    }
    /**
     * Get the weight of a single morph target.
     *
     * @param target Index of the morph target.
     * @returns The weight.
     *
     * @since 1.2.0
     */
    getMorphTargetWeight(target) {
        const count = this.morphTargets?.count ?? 0;
        if (target >= count) {
            throw new Error(`Index ${target} is out of bounds for ${count} targets`);
        }
        return this.engine.wasm._wl_mesh_component_get_morph_target_weight(this._id, target);
    }
    /**
     * Set morph target weights for this mesh component.
     *
     * @param weights Array of new weights, expected to have at least as many
     *     elements as {@link MorphTargets.count}.
     *
     * @since 1.2.0
     */
    setMorphTargetWeights(weights) {
        const count = this.morphTargets?.count ?? 0;
        if (weights.length !== count) {
            throw new Error(`Expected ${count} weights but got ${weights.length}`);
        }
        const wasm = this.engine.wasm;
        wasm._tempMemFloat.set(weights);
        wasm._wl_mesh_component_set_morph_target_weights(this._id, wasm._tempMem, weights.length);
    }
    /**
     * Set the weight of a single morph target.
     *
     * @param target Index of the morph target.
     * @param weight The new weight.
     *
     * ## Usage
     *
     * ```js
     * const mesh = object.getComponent('mesh');
     * const mouthTarget = mesh.morphTargets.getTargetIndex('mouth');
     * mesh.setMorphTargetWeight(mouthTarget, 0.5);
     * ```
     *
     * @since 1.2.0
     */
    setMorphTargetWeight(target, weight) {
        const count = this.morphTargets?.count ?? 0;
        if (target >= count) {
            throw new Error(`Index ${target} is out of bounds for ${count} targets`);
        }
        this.engine.wasm._wl_mesh_component_set_morph_target_weight(this._id, target, weight);
    }
}
__decorate([
    nativeProperty()
], MeshComponent.prototype, "material", null);
__decorate([
    nativeProperty()
], MeshComponent.prototype, "mesh", null);
__decorate([
    nativeProperty()
], MeshComponent.prototype, "skin", null);
__decorate([
    nativeProperty()
], MeshComponent.prototype, "morphTargets", null);
__decorate([
    nativeProperty()
], MeshComponent.prototype, "morphTargetWeights", null);
/**
 * Enum for Physics axes locking
 *
 * See {@link PhysXComponent.angularLockAxis} and {@link PhysXComponent.linearLockAxis}.
 */
export var LockAxis;
(function (LockAxis) {
    /**
     * No axis selected.
     */
    LockAxis[LockAxis["None"] = 0] = "None";
    /**
     * **X axis**:
     */
    LockAxis[LockAxis["X"] = 1] = "X";
    /**
     * **Y axis**:
     */
    LockAxis[LockAxis["Y"] = 2] = "Y";
    /**
     * **Z axis**:
     */
    LockAxis[LockAxis["Z"] = 4] = "Z";
})(LockAxis || (LockAxis = {}));
/**
 * Native physx rigid body component.
 *
 * Provides access to a native mesh component instance.
 * Only available when using physx enabled runtime, see "Project Settings > Runtime".
 */
export class PhysXComponent extends Component {
    /** @override */
    static TypeName = 'physx';
    getTranslationOffset(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        wasm._wl_physx_component_get_offsetTranslation(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    getRotationOffset(out = new Float32Array(4)) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_physx_component_get_offsetTransform(this._id) >> 2; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        out[3] = wasm.HEAPF32[ptr + 3];
        return out;
    }
    getExtents(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        const ptr = wasm._wl_physx_component_get_extents(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        return out;
    }
    getLinearVelocity(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        const tempMemFloat = wasm._tempMemFloat;
        wasm._wl_physx_component_get_linearVelocity(this._id, wasm._tempMem); /* Align F32 */
        out[0] = tempMemFloat[0];
        out[1] = tempMemFloat[1];
        out[2] = tempMemFloat[2];
        return out;
    }
    getAngularVelocity(out = new Float32Array(3)) {
        const wasm = this.engine.wasm;
        const tempMemFloat = wasm._tempMemFloat;
        wasm._wl_physx_component_get_angularVelocity(this._id, wasm._tempMem); /* Align F32 */
        out[0] = tempMemFloat[0];
        out[1] = tempMemFloat[1];
        out[2] = tempMemFloat[2];
        return out;
    }
    /**
     * Set whether this rigid body is static.
     *
     * Setting this property only takes effect once the component
     * switches from inactive to active.
     *
     * @param b Whether the rigid body should be static.
     */
    set static(b) {
        this.engine.wasm._wl_physx_component_set_static(this._id, b);
    }
    /**
     * Whether this rigid body is static.
     *
     * This property returns whether the rigid body is *effectively*
     * static. If static property was set while the rigid body was
     * active, it will not take effect until the rigid body is set
     * inactive and active again. Until the component is set inactive,
     * this getter will return whether the rigid body is actually
     * static.
     */
    get static() {
        return !!this.engine.wasm._wl_physx_component_get_static(this._id);
    }
    /**
     * Equivalent to {@link PhysXComponent.getTranslationOffset}.
     *
     * Gives a quick view of the offset in a debugger.
     *
     * @note Prefer to use {@link PhysXComponent.getTranslationOffset} for performance.
     *
     * @since 1.1.1
     */
    get translationOffset() {
        return this.getTranslationOffset();
    }
    /**
     * Set the offset translation.
     *
     * The array must be a vector of at least **3** elements.
     *
     * @note The component must be re-activated to apply the change.
     *
     * @since 1.1.1
     */
    set translationOffset(offset) {
        const wasm = this.engine.wasm;
        wasm._wl_physx_component_set_offsetTranslation(this._id, offset[0], offset[1], offset[2]);
    }
    /**
     * Equivalent to {@link PhysXComponent.getRotationOffset}.
     *
     * Gives a quick view of the offset in a debugger.
     *
     * @note Prefer to use {@link PhysXComponent.getRotationOffset} for performance.
     *
     * @since 1.1.1
     */
    get rotationOffset() {
        return this.getRotationOffset();
    }
    /**
     * Set the offset rotation.
     *
     * The array must be a quaternion of at least **4** elements.
     *
     * @note The component must be re-activated to apply the change.
     *
     * @since 1.1.1
     */
    set rotationOffset(offset) {
        const wasm = this.engine.wasm;
        wasm._wl_physx_component_set_offsetRotation(this._id, offset[0], offset[1], offset[2], offset[3]);
    }
    /**
     * Set whether this rigid body is kinematic.
     *
     * @param b Whether the rigid body should be kinematic.
     */
    set kinematic(b) {
        this.engine.wasm._wl_physx_component_set_kinematic(this._id, b);
    }
    /**
     * Whether this rigid body is kinematic.
     */
    get kinematic() {
        return !!this.engine.wasm._wl_physx_component_get_kinematic(this._id);
    }
    /**
     * Set whether this rigid body's gravity is enabled.
     *
     * @param b Whether the rigid body's gravity should be enabled.
     */
    set gravity(b) {
        this.engine.wasm._wl_physx_component_set_gravity(this._id, b);
    }
    /**
     * Whether this rigid body's gravity flag is enabled.
     */
    get gravity() {
        return !!this.engine.wasm._wl_physx_component_get_gravity(this._id);
    }
    /**
     * Set whether this rigid body's simulate flag is enabled.
     *
     * @param b Whether the rigid body's simulate flag should be enabled.
     */
    set simulate(b) {
        this.engine.wasm._wl_physx_component_set_simulate(this._id, b);
    }
    /**
     * Whether this rigid body's simulate flag is enabled.
     */
    get simulate() {
        return !!this.engine.wasm._wl_physx_component_get_simulate(this._id);
    }
    /**
     * Set whether to allow simulation of this rigid body.
     *
     * {@link allowSimulation} and {@link trigger} can not be enabled at the
     * same time. Enabling {@link allowSimulation} while {@link trigger} is enabled
     * will disable {@link trigger}.
     *
     * @param b Whether to allow simulation of this rigid body.
     */
    set allowSimulation(b) {
        this.engine.wasm._wl_physx_component_set_allowSimulation(this._id, b);
    }
    /**
     * Whether to allow simulation of this rigid body.
     */
    get allowSimulation() {
        return !!this.engine.wasm._wl_physx_component_get_allowSimulation(this._id);
    }
    /**
     * Set whether this rigid body may be queried in ray casts.
     *
     * @param b Whether this rigid body may be queried in ray casts.
     */
    set allowQuery(b) {
        this.engine.wasm._wl_physx_component_set_allowQuery(this._id, b);
    }
    /**
     * Whether this rigid body may be queried in ray casts.
     */
    get allowQuery() {
        return !!this.engine.wasm._wl_physx_component_get_allowQuery(this._id);
    }
    /**
     * Set whether this physics body is a trigger.
     *
     * {@link allowSimulation} and {@link trigger} can not be enabled at the
     * same time. Enabling trigger while {@link allowSimulation} is enabled,
     * will disable {@link allowSimulation}.
     *
     * @param b Whether this physics body is a trigger.
     */
    set trigger(b) {
        this.engine.wasm._wl_physx_component_set_trigger(this._id, b);
    }
    /**
     * Whether this physics body is a trigger.
     */
    get trigger() {
        return !!this.engine.wasm._wl_physx_component_get_trigger(this._id);
    }
    /**
     * Set the shape for collision detection.
     *
     * @param s New shape.
     * @since 0.8.5
     */
    set shape(s) {
        this.engine.wasm._wl_physx_component_set_shape(this._id, s);
    }
    /** The shape for collision detection. */
    get shape() {
        return this.engine.wasm._wl_physx_component_get_shape(this._id);
    }
    /**
     * Set additional data for the shape.
     *
     * Retrieved only from {@link PhysXComponent#shapeData}.
     * @since 0.8.10
     */
    set shapeData(d) {
        if (d == null || !isMeshShape(this.shape))
            return;
        this.engine.wasm._wl_physx_component_set_shape_data(this._id, d.index);
    }
    /**
     * Additional data for the shape.
     *
     * `null` for {@link Shape} values: `None`, `Sphere`, `Capsule`, `Box`, `Plane`.
     * `{index: n}` for `TriangleMesh` and `ConvexHull`.
     *
     * This data is currently only for passing onto or creating other {@link PhysXComponent}.
     * @since 0.8.10
     */
    get shapeData() {
        if (!isMeshShape(this.shape))
            return null;
        return {
            index: this.engine.wasm._wl_physx_component_get_shape_data(this._id),
        };
    }
    /**
     * Set the shape extents for collision detection.
     *
     * @param e New extents for the shape.
     * @since 0.8.5
     */
    set extents(e) {
        this.extents.set(e);
    }
    /**
     * Equivalent to {@link PhysXComponent.getExtents}.
     *
     * @note Prefer to use {@link PhysXComponent.getExtents} for performance.
     */
    get extents() {
        const wasm = this.engine.wasm;
        /** @todo: Break at 2.0.0. Do not allow modifying memory in-place. */
        const ptr = wasm._wl_physx_component_get_extents(this._id);
        return new Float32Array(wasm.HEAPF32.buffer, ptr, 3);
    }
    /**
     * Get staticFriction.
     */
    get staticFriction() {
        return this.engine.wasm._wl_physx_component_get_staticFriction(this._id);
    }
    /**
     * Set staticFriction.
     * @param v New staticFriction.
     */
    set staticFriction(v) {
        this.engine.wasm._wl_physx_component_set_staticFriction(this._id, v);
    }
    /**
     * Get dynamicFriction.
     */
    get dynamicFriction() {
        return this.engine.wasm._wl_physx_component_get_dynamicFriction(this._id);
    }
    /**
     * Set dynamicFriction
     * @param v New dynamicDamping.
     */
    set dynamicFriction(v) {
        this.engine.wasm._wl_physx_component_set_dynamicFriction(this._id, v);
    }
    /**
     * Get bounciness.
     * @since 0.9.0
     */
    get bounciness() {
        return this.engine.wasm._wl_physx_component_get_bounciness(this._id);
    }
    /**
     * Set bounciness.
     * @param v New bounciness.
     * @since 0.9.0
     */
    set bounciness(v) {
        this.engine.wasm._wl_physx_component_set_bounciness(this._id, v);
    }
    /**
     * Get linearDamping/
     */
    get linearDamping() {
        return this.engine.wasm._wl_physx_component_get_linearDamping(this._id);
    }
    /**
     * Set linearDamping.
     * @param v New linearDamping.
     */
    set linearDamping(v) {
        this.engine.wasm._wl_physx_component_set_linearDamping(this._id, v);
    }
    /** Get angularDamping. */
    get angularDamping() {
        return this.engine.wasm._wl_physx_component_get_angularDamping(this._id);
    }
    /**
     * Set angularDamping.
     * @param v New angularDamping.
     */
    set angularDamping(v) {
        this.engine.wasm._wl_physx_component_set_angularDamping(this._id, v);
    }
    /**
     * Set linear velocity.
     *
     * [PhysX Manual - "Velocity"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#velocity)
     *
     * Has no effect, if the component is not active.
     *
     * @param v New linear velocity.
     */
    set linearVelocity(v) {
        this.engine.wasm._wl_physx_component_set_linearVelocity(this._id, v[0], v[1], v[2]);
    }
    /**
     * Equivalent to {@link PhysXComponent.getLinearVelocity}.
     *
     * @note Prefer to use {@link PhysXComponent.getLinearVelocity} for performance.
     */
    get linearVelocity() {
        /** @todo: Break at 2.0.0. Do not allow modifying memory in-place. */
        const wasm = this.engine.wasm;
        wasm._wl_physx_component_get_linearVelocity(this._id, wasm._tempMem);
        return new Float32Array(wasm.HEAPF32.buffer, wasm._tempMem, 3);
    }
    /**
     * Set angular velocity
     *
     * [PhysX Manual - "Velocity"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#velocity)
     *
     * Has no effect, if the component is not active.
     *
     * @param v New angular velocity
     */
    set angularVelocity(v) {
        this.engine.wasm._wl_physx_component_set_angularVelocity(this._id, v[0], v[1], v[2]);
    }
    /**
     * Equivalent to {@link PhysXComponent.getAngularVelocity}.
     *
     * @note Prefer to use {@link PhysXComponent.getAngularVelocity} for performance.
     */
    get angularVelocity() {
        /** @todo: Break at 2.0.0. Do not allow modifying memory in-place. */
        const wasm = this.engine.wasm;
        wasm._wl_physx_component_get_angularVelocity(this._id, wasm._tempMem);
        return new Float32Array(wasm.HEAPF32.buffer, wasm._tempMem, 3);
    }
    /**
     * Set the components groups mask.
     *
     * @param flags New flags that need to be set.
     */
    set groupsMask(flags) {
        this.engine.wasm._wl_physx_component_set_groupsMask(this._id, flags);
    }
    /**
     * Get the components groups mask flags.
     *
     * Each bit represents membership to group, see example.
     *
     * ```js
     * // Assign c to group 2
     * c.groupsMask = (1 << 2);
     *
     * // Assign c to group 0
     * c.groupsMask  = (1 << 0);
     *
     * // Assign c to group 0 and 2
     * c.groupsMask = (1 << 0) | (1 << 2);
     *
     * (c.groupsMask & (1 << 2)) != 0; // true
     * (c.groupsMask & (1 << 7)) != 0; // false
     * ```
     */
    get groupsMask() {
        return this.engine.wasm._wl_physx_component_get_groupsMask(this._id);
    }
    /**
     * Set the components blocks mask.
     *
     * @param flags New flags that need to be set.
     */
    set blocksMask(flags) {
        this.engine.wasm._wl_physx_component_set_blocksMask(this._id, flags);
    }
    /**
     * Get the components blocks mask flags.
     *
     * Each bit represents membership to the block, see example.
     *
     * ```js
     * // Block overlap with any objects in group 2
     * c.blocksMask = (1 << 2);
     *
     * // Block overlap with any objects in group 0
     * c.blocksMask  = (1 << 0)
     *
     * // Block overlap with any objects in group 0 and 2
     * c.blocksMask = (1 << 0) | (1 << 2);
     *
     * (c.blocksMask & (1 << 2)) != 0; // true
     * (c.blocksMask & (1 << 7)) != 0; // false
     * ```
     */
    get blocksMask() {
        return this.engine.wasm._wl_physx_component_get_blocksMask(this._id);
    }
    /**
     * Set axes to lock for linear velocity.
     *
     * @param lock The Axis that needs to be set.
     *
     * Combine flags with Bitwise OR:
     *
     * ```js
     * body.linearLockAxis = LockAxis.X | LockAxis.Y; // x and y set
     * body.linearLockAxis = LockAxis.X; // y unset
     * ```
     *
     * @note This has no effect if the component is static.
     */
    set linearLockAxis(lock) {
        this.engine.wasm._wl_physx_component_set_linearLockAxis(this._id, lock);
    }
    /**
     * Get the linear lock axes flags.
     *
     * To get the state of a specific flag, Bitwise AND with the LockAxis needed.
     *
     * ```js
     * if(body.linearLockAxis & LockAxis.Y) {
     *     console.log("The Y flag was set!");
     * }
     * ```
     *
     * @return axes that are currently locked for linear movement.
     */
    get linearLockAxis() {
        return this.engine.wasm._wl_physx_component_get_linearLockAxis(this._id);
    }
    /**
     * Set axes to lock for angular velocity.
     *
     * @param lock The Axis that needs to be set.
     *
     * ```js
     * body.angularLockAxis = LockAxis.X | LockAxis.Y; // x and y set
     * body.angularLockAxis = LockAxis.X; // y unset
     * ```
     *
     * @note This has no effect if the component is static.
     */
    set angularLockAxis(lock) {
        this.engine.wasm._wl_physx_component_set_angularLockAxis(this._id, lock);
    }
    /**
     * Get the angular lock axes flags.
     *
     * To get the state of a specific flag, Bitwise AND with the LockAxis needed:
     *
     * ```js
     * if(body.angularLockAxis & LockAxis.Y) {
     *     console.log("The Y flag was set!");
     * }
     * ```
     *
     * @return axes that are currently locked for angular movement.
     */
    get angularLockAxis() {
        return this.engine.wasm._wl_physx_component_get_angularLockAxis(this._id);
    }
    /**
     * Set mass.
     *
     * [PhysX Manual - "Mass Properties"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#mass-properties)
     *
     * @param m New mass.
     */
    set mass(m) {
        this.engine.wasm._wl_physx_component_set_mass(this._id, m);
    }
    /** Mass */
    get mass() {
        return this.engine.wasm._wl_physx_component_get_mass(this._id);
    }
    /**
     * Set mass space interia tensor.
     *
     * [PhysX Manual - "Mass Properties"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#mass-properties)
     *
     * Has no effect, if the component is not active.
     *
     * @param v New mass space interatia tensor.
     */
    set massSpaceInteriaTensor(v) {
        this.engine.wasm._wl_physx_component_set_massSpaceInertiaTensor(this._id, v[0], v[1], v[2]);
    }
    /**
     * Set the rigid body to sleep upon activation.
     *
     * When asleep, the rigid body will not be simulated until the next contact.
     *
     * @param flag `true` to sleep upon activation.
     *
     * @since 1.1.5
     */
    set sleepOnActivate(flag) {
        this.engine.wasm._wl_physx_component_set_sleepOnActivate(this._id, flag);
    }
    /**
     * `true` if the rigid body is set to sleep upon activation, `false` otherwise.
     *
     * @since 1.1.5
     */
    get sleepOnActivate() {
        return !!this.engine.wasm._wl_physx_component_get_sleepOnActivate(this._id);
    }
    /**
     * Apply a force.
     *
     * [PhysX Manual - "Applying Forces and Torques"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#applying-forces-and-torques)
     *
     * Has no effect, if the component is not active.
     *
     * @param f Force vector.
     * @param m Force mode, see {@link ForceMode}, default `Force`.
     * @param localForce Whether the force vector is in local space, default `false`.
     * @param p Position to apply force at, default is center of mass.
     * @param local Whether position is in local space, default `false`.
     */
    addForce(f, m = ForceMode.Force, localForce = false, p, local = false) {
        const wasm = this.engine.wasm;
        if (!p) {
            wasm._wl_physx_component_addForce(this._id, f[0], f[1], f[2], m, localForce);
            return;
        }
        wasm._wl_physx_component_addForceAt(this._id, f[0], f[1], f[2], m, localForce, p[0], p[1], p[2], local);
    }
    /**
     * Apply torque.
     *
     * [PhysX Manual - "Applying Forces and Torques"](https://gameworksdocs.nvidia.com/PhysX/4.1/documentation/physxguide/Manual/RigidBodyDynamics.html#applying-forces-and-torques)
     *
     * Has no effect, if the component is not active.
     *
     * @param f Force vector.
     * @param m Force mode, see {@link ForceMode}, default `Force`.
     */
    addTorque(f, m = ForceMode.Force) {
        this.engine.wasm._wl_physx_component_addTorque(this._id, f[0], f[1], f[2], m);
    }
    /**
     * Add on collision callback.
     *
     * @param callback Function to call when this rigid body (un)collides with any other.
     *
     * ```js
     *  let rigidBody = this.object.getComponent('physx');
     *  rigidBody.onCollision(function(type, other) {
     *      // Ignore uncollides
     *      if(type == CollisionEventType.TouchLost) return;
     *
     *      // Take damage on collision with enemies
     *      if(other.object.name.startsWith("enemy-")) {
     *          this.applyDamage(10);
     *      }
     *  }.bind(this));
     * ```
     *
     * @returns Id of the new callback for use with {@link PhysXComponent#removeCollisionCallback}.
     */
    onCollision(callback) {
        return this.onCollisionWith(this, callback);
    }
    /**
     * Add filtered on collision callback.
     *
     * @param otherComp Component for which callbacks will
     *        be triggered. If you pass this component, the method is equivalent to.
     *        {@link PhysXComponent#onCollision}.
     * @param callback Function to call when this rigid body
     *        (un)collides with `otherComp`.
     * @returns Id of the new callback for use with {@link PhysXComponent#removeCollisionCallback}.
     */
    onCollisionWith(otherComp, callback) {
        const physics = this.engine.physics;
        physics._callbacks[this._id] = physics._callbacks[this._id] || [];
        physics._callbacks[this._id].push(callback);
        return this.engine.wasm._wl_physx_component_addCallback(this._id, otherComp._id || this._id);
    }
    /**
     * Remove a collision callback added with {@link PhysXComponent#onCollision} or {@link PhysXComponent#onCollisionWith}.
     *
     * @param callbackId Callback id as returned by {@link PhysXComponent#onCollision} or {@link PhysXComponent#onCollisionWith}.
     * @throws When the callback does not belong to the component.
     * @throws When the callback does not exist.
     */
    removeCollisionCallback(callbackId) {
        const physics = this.engine.physics;
        const r = this.engine.wasm._wl_physx_component_removeCallback(this._id, callbackId);
        /* r is the amount of object to remove from the end of the
         * callbacks array for this object */
        if (r)
            physics._callbacks[this._id].splice(-r);
    }
}
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "static", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "translationOffset", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "rotationOffset", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "kinematic", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "gravity", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "simulate", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "allowSimulation", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "allowQuery", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "trigger", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "shape", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "shapeData", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "extents", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "staticFriction", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "dynamicFriction", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "bounciness", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "linearDamping", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "angularDamping", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "linearVelocity", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "angularVelocity", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "groupsMask", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "blocksMask", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "linearLockAxis", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "angularLockAxis", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "mass", null);
__decorate([
    nativeProperty()
], PhysXComponent.prototype, "sleepOnActivate", null);
/**
 * Access to the physics scene
 */
export class Physics {
    /**
     * @hidden
     *
     * **Note**: This is public to emulate a `friend` accessor.
     */
    _callbacks;
    /**
     * Hit.
     * @hidden
     */
    _hit;
    /**
     * Wonderland Engine instance
     * @hidden
     */
    _engine;
    /**
     * Ray Hit
     * @hidden
     */
    _rayHit;
    constructor(engine) {
        this._engine = engine;
        this._rayHit = engine.wasm._malloc(4 * (3 * 4 + 3 * 4 + 4 + 2) + 4);
        this._hit = new RayHit(engine.scene, this._rayHit);
        this._callbacks = {};
    }
    /**
     * Cast a ray through the scene and find intersecting physics components.
     *
     * The resulting ray hit will contain **up to 4** closest ray hits,
     * sorted by increasing distance.
     *
     * Example:
     *
     * ```js
     * const hit = engine.physics.rayCast(
     *     [0, 0, 0],
     *     [0, 0, 1],
     *     1 << 0 | 1 << 4, // Only check against physics components in groups 0 and 4
     *     25
     * );
     * if (hit.hitCount > 0) {
     *     const locations = hit.getLocations();
     *     console.log(`Object hit at: ${locations[0][0]}, ${locations[0][1]}, ${locations[0][2]}`);
     * }
     * ```
     *
     * @param o Ray origin.
     * @param d Ray direction.
     * @param groupMask Bitmask of physics groups to filter by: only objects
     *        that are part of given groups are considered for the raycast.
     * @param maxDistance Maximum **inclusive** hit distance. Defaults to `100`.
     *
     * @returns The {@link RayHit} instance, cached by this class.
     *
     * @note The returned {@link RayHit} object is owned by the {@link Physics}
     *       instance and will be reused with the next {@link Physics#rayCast} call.
     */
    rayCast(o, d, groupMask, maxDistance = 100.0) {
        const scene = this._engine.scene._index;
        this._engine.wasm._wl_physx_ray_cast(scene, o[0], o[1], o[2], d[0], d[1], d[2], groupMask, maxDistance, this._rayHit);
        return this._hit;
    }
    /** Hosting engine instance. */
    get engine() {
        return this._engine;
    }
}
/**
 * Mesh index type.
 */
export var MeshIndexType;
(function (MeshIndexType) {
    /** Single byte mesh index, range 0-255 */
    MeshIndexType[MeshIndexType["UnsignedByte"] = 1] = "UnsignedByte";
    /** Two byte mesh index, range 0-65535 */
    MeshIndexType[MeshIndexType["UnsignedShort"] = 2] = "UnsignedShort";
    /** Four byte mesh index, range 0-4294967295 */
    MeshIndexType[MeshIndexType["UnsignedInt"] = 4] = "UnsignedInt";
})(MeshIndexType || (MeshIndexType = {}));
/**
 * Mesh skinning type.
 */
export var MeshSkinningType;
(function (MeshSkinningType) {
    /** Not skinned */
    MeshSkinningType[MeshSkinningType["None"] = 0] = "None";
    /** Skinned, 4 joints/weight per vertex */
    MeshSkinningType[MeshSkinningType["FourJoints"] = 1] = "FourJoints";
    /** Skinned, 8 joints/weight per vertex */
    MeshSkinningType[MeshSkinningType["EightJoints"] = 2] = "EightJoints";
})(MeshSkinningType || (MeshSkinningType = {}));
/**
 * Wrapper around a native mesh data.
 *
 * For more information about how to create meshes, have a look at the
 * {@link MeshManager} class.
 *
 * #### Update
 *
 * To modify a mesh, you get access to a {@link MeshAttributeAccessor} that
 * allows you to modify the content of the buffers:
 *
 * Usage:
 *
 * ```js
 * const mesh = engine.es.create({vertexCount: 3, indexData: [0, 1, 2]});
 * const positions = mesh.attribute(MeshAttribute.Position);
 * ...
 * ```
 *
 * @note Meshes are **per-engine**, they can thus be shared by multiple scenes.
 */
export class Mesh extends Resource {
    /**
     * @deprecated Use {@link MeshManager.create} instead, accessible via {@link WonderlandEngine.meshes}:
     *
     * ```js
     * const mesh = engine.meshes.create({vertexCount: 3, indexData: [0, 1, 2]});
     * ...
     * mesh.update();
     * ```
     */
    constructor(engine, params) {
        if (!isNumber(params)) {
            const mesh = engine.meshes.create(params);
            /* `super()` must be called as stated in the specification */
            super(engine, mesh._index);
            return mesh;
        }
        super(engine, params);
    }
    /** Number of vertices in this mesh. */
    get vertexCount() {
        return this.engine.wasm._wl_mesh_get_vertexCount(this._id);
    }
    /** Index data (read-only) or `null` if the mesh is not indexed. */
    get indexData() {
        const wasm = this.engine.wasm;
        const tempMem = wasm._tempMem;
        const ptr = wasm._wl_mesh_get_indexData(this._id, tempMem, tempMem + 4);
        if (ptr === null)
            return null;
        const indexCount = wasm.HEAPU32[tempMem / 4];
        const indexSize = wasm.HEAPU32[tempMem / 4 + 1];
        switch (indexSize) {
            case MeshIndexType.UnsignedByte:
                return new Uint8Array(wasm.HEAPU8.buffer, ptr, indexCount);
            case MeshIndexType.UnsignedShort:
                return new Uint16Array(wasm.HEAPU16.buffer, ptr, indexCount);
            case MeshIndexType.UnsignedInt:
                return new Uint32Array(wasm.HEAPU32.buffer, ptr, indexCount);
        }
        return null;
    }
    /**
     * Apply changes to {@link attribute | vertex attributes}.
     *
     * Uploads the updated vertex attributes to the GPU and updates the bounding
     * sphere to match the new vertex positions.
     *
     * Since this is an expensive operation, call it only once you have performed
     * all modifications on a mesh and avoid calling if you did not perform any
     * modifications at all.
     */
    update() {
        this.engine.wasm._wl_mesh_update(this._id);
    }
    getBoundingSphere(out = new Float32Array(4)) {
        const wasm = this.engine.wasm;
        this.engine.wasm._wl_mesh_get_boundingSphere(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        out[3] = wasm._tempMemFloat[3];
        return out;
    }
    attribute(attr) {
        if (typeof attr != 'number')
            throw new TypeError('Expected number, but got ' + typeof attr);
        const wasm = this.engine.wasm;
        const tempMemUint32 = wasm._tempMemUint32;
        wasm._wl_mesh_get_attribute(this._id, attr, wasm._tempMem);
        if (tempMemUint32[0] == 255)
            return null;
        const arraySize = tempMemUint32[5];
        return new MeshAttributeAccessor(this.engine, {
            attribute: tempMemUint32[0],
            offset: tempMemUint32[1],
            stride: tempMemUint32[2],
            formatSize: tempMemUint32[3],
            componentCount: tempMemUint32[4],
            /* The WASM API returns `0` for a scalar value. We clamp it to 1 as we strictly use it as a multiplier for get/set operations */
            arraySize: arraySize ? arraySize : 1,
            length: this.vertexCount,
            bufferType: (attr !== MeshAttribute.JointId
                ? Float32Array
                : Uint16Array),
        });
    }
    /**
     * Destroy and free the meshes memory.
     *
     * It is best practice to set the mesh variable to `null` after calling
     * destroy to prevent accidental use:
     *
     * ```js
     *   mesh.destroy();
     *   mesh = null;
     * ```
     *
     * Accessing the mesh after destruction behaves like accessing an empty
     * mesh.
     *
     * @since 0.9.0
     */
    destroy() {
        this.engine.wasm._wl_mesh_destroy(this._id);
        this.engine.meshes._destroy(this);
    }
    toString() {
        if (this.isDestroyed) {
            return 'Mesh(destroyed)';
        }
        return `Mesh(${this._index})`;
    }
}
/**
 * An iterator over a mesh vertex attribute.
 *
 * Usage:
 *
 * ```js
 * const mesh = this.object.getComponent('mesh').mesh;
 * const positions = mesh.attribute(MeshAttribute.Position);
 *
 * // Equivalent to `new Float32Array(3)`.
 * const temp = positions.createArray();
 *
 * for(let i = 0; i < positions.length; ++i) {
 *     // `pos` will reference `temp` and thereby not allocate additional
 *     // memory, which would cause a perf spike when collected.
 *     const pos = positions.get(i, temp);
 *     // Scale position by 2 on x-axis only.
 *     pos[0] *= 2.0;
 *     positions.set(i, pos);
 * }
 * // We're done modifying, tell the engine to move vertex data to the GPU.
 * mesh.update();
 * ```
 */
export class MeshAttributeAccessor {
    /** Max number of elements. */
    length = 0;
    /** Wonderland Engine instance. @hidden */
    _engine;
    /** Attribute index. @hidden */
    _attribute = -1;
    /** Attribute offset. @hidden */
    _offset = 0;
    /** Attribute stride. @hidden */
    _stride = 0;
    /** Format size native enum. @hidden */
    _formatSize = 0;
    /** Number of components per vertex. @hidden */
    _componentCount = 0;
    /** Number of values per vertex. @hidden */
    _arraySize = 1;
    /**
     * Class to instantiate an ArrayBuffer to get/set values.
     */
    _bufferType;
    /**
     * Function to allocate temporary WASM memory. It is cached in the accessor to avoid
     * conditionals during get/set.
     */
    _tempBufferGetter;
    /**
     * Create a new instance.
     *
     * @note Please use {@link Mesh.attribute} to create a new instance.
     *
     * @param options Contains information about how to read the data.
     * @note Do not use this constructor. Instead, please use the {@link Mesh.attribute} method.
     *
     * @hidden
     */
    constructor(engine, options) {
        this._engine = engine;
        const wasm = this._engine.wasm;
        this._attribute = options.attribute;
        this._offset = options.offset;
        this._stride = options.stride;
        this._formatSize = options.formatSize;
        this._componentCount = options.componentCount;
        this._arraySize = options.arraySize;
        this._bufferType = options.bufferType;
        this.length = options.length;
        this._tempBufferGetter = (this._bufferType === Float32Array
            ? wasm.getTempBufferF32.bind(wasm)
            : wasm.getTempBufferU16.bind(wasm));
    }
    /**
     * Create a new TypedArray to hold this attribute's values.
     *
     * This method is useful to create a view to hold the data to
     * pass to {@link get} and {@link set}
     *
     * Example:
     *
     * ```js
     * const vertexCount = 4;
     * const positionAttribute = mesh.attribute(MeshAttribute.Position);
     *
     * // A position has 3 floats per vertex. Thus, positions has length 3 * 4.
     * const positions = positionAttribute.createArray(vertexCount);
     * ```
     *
     * @param count The number of **vertices** expected.
     * @returns A TypedArray with the appropriate format to access the data
     */
    createArray(count = 1) {
        count = count > this.length ? this.length : count;
        return new this._bufferType(count * this._componentCount * this._arraySize);
    }
    get(index, out = this.createArray()) {
        if (out.length % this._componentCount !== 0) {
            throw new Error(`out.length, ${out.length} is not a multiple of the attribute vector components, ${this._componentCount}`);
        }
        const dest = this._tempBufferGetter(out.length);
        const elementSize = this._bufferType.BYTES_PER_ELEMENT;
        const destSize = elementSize * out.length;
        const srcFormatSize = this._formatSize * this._arraySize;
        const destFormatSize = this._componentCount * elementSize * this._arraySize;
        this._engine.wasm._wl_mesh_get_attribute_values(this._attribute, srcFormatSize, this._offset + index * this._stride, this._stride, destFormatSize, dest.byteOffset, destSize);
        for (let i = 0; i < out.length; ++i)
            out[i] = dest[i];
        return out;
    }
    /**
     * Set attribute element.
     *
     * @param i Index
     * @param v Value to set the element to
     *
     * `v.length` needs to be a multiple of the attributes component count, see
     * {@link MeshAttribute}. If `v.length` is more than one multiple, it will be
     * filled with the next n attribute elements, which can reduce overhead
     * of this call.
     *
     * @returns Reference to self (for method chaining)
     */
    set(i, v) {
        if (v.length % this._componentCount !== 0)
            throw new Error(`out.length, ${v.length} is not a multiple of the attribute vector components, ${this._componentCount}`);
        const elementSize = this._bufferType.BYTES_PER_ELEMENT;
        const srcSize = elementSize * v.length;
        const srcFormatSize = this._componentCount * elementSize * this._arraySize;
        const destFormatSize = this._formatSize * this._arraySize;
        const wasm = this._engine.wasm;
        /* Unless we are already working with data from WASM heap, we
         * need to copy into temporary memory. */
        if (v.buffer != wasm.HEAPU8.buffer) {
            const dest = this._tempBufferGetter(v.length);
            dest.set(v);
            v = dest;
        }
        wasm._wl_mesh_set_attribute_values(this._attribute, srcFormatSize, v.byteOffset, srcSize, destFormatSize, this._offset + i * this._stride, this._stride);
        return this;
    }
    /** Hosting engine instance. */
    get engine() {
        return this._engine;
    }
}
/**
 * Wrapper around a native font resource.
 *
 * @note Fonts are **per-engine**, they can thus be shared by multiple scenes.
 *
 * @since 1.2.0
 */
export class Font extends Resource {
    /** Em height in object space. Equivalent to line height. */
    get emHeight() {
        return this.engine.wasm._wl_font_get_emHeight(this._id);
    }
    /**
     * Cap height in object space. This is the typical height of capital
     * letters. Can be 0 if not defined by the font.
     */
    get capHeight() {
        return this.engine.wasm._wl_font_get_capHeight(this._id);
    }
    /**
     * X height in object space. This is the typical height of lowercase
     * letters. Can be 0 if not defined by the font.
     */
    get xHeight() {
        return this.engine.wasm._wl_font_get_xHeight(this._id);
    }
    /**
     * Outline size. This is the factor by which characters are expanded to
     * create the outline effect. Returns 0 if the font was compiled without
     * an outline.
     *
     * @since 1.2.1
     */
    get outlineSize() {
        return this.engine.wasm._wl_font_get_outlineSize(this._id);
    }
}
/** Temporary canvas */
let temp2d = null;
/**
 * Wrapper around a native texture data.
 *
 * For more information about how to create meshes, have a look at the
 * {@link TextureManager} class.
 *
 * @note Textures are **per-engine**, they can thus be shared by multiple scenes.
 */
export class Texture extends Resource {
    /**
     * @deprecated Use {@link TextureManager.create} instead, accessible via
     * {@link WonderlandEngine.textures}:
     *
     * ```js
     * const image = new Image();
     * image.onload = () => {
     *     const texture = engine.textures.create(image);
     * };
     * ```
     */
    constructor(engine, param) {
        if (isImageLike(param)) {
            const texture = engine.textures.create(param);
            /* `super()` must be called as stated in the specification */
            super(engine, texture._index);
            return texture;
        }
        super(engine, param);
    }
    /**
     * Whether this texture is valid
     *
     * @deprecated Use {@link SceneResource#isDestroyed} instead.
     */
    get valid() {
        return !this.isDestroyed;
    }
    /**
     * Index in this manager.
     *
     * @deprecated Use {@link Texture.index} instead.
     */
    get id() {
        return this.index;
    }
    /** Update the texture to match the HTML element (e.g. reflect the current frame of a video). */
    update() {
        const image = this._imageIndex;
        if (!this.valid || !image)
            return;
        this.engine.wasm._wl_renderer_updateImage(image);
    }
    /** Width of the texture. */
    get width() {
        /* HTML textures should be read directly from js, since the C++ could be
         * one frame out-of-sync on the size of the element itself. */
        const element = this.htmlElement;
        if (element)
            return element.width;
        const wasm = this.engine.wasm;
        wasm._wl_image_size(this._imageIndex, wasm._tempMem);
        return wasm._tempMemUint32[0];
    }
    /** Height of the texture. */
    get height() {
        /* HTML textures should be read directly from js, since the C++ could be
         * one frame out-of-sync on the size of the element itself. */
        const element = this.htmlElement;
        if (element)
            return element.height;
        const wasm = this.engine.wasm;
        wasm._wl_image_size(this._imageIndex, wasm._tempMem);
        return wasm._tempMemUint32[1];
    }
    /**
     * Returns the html element associated to this texture.
     *
     * @note This accessor will return `null` if the image is compressed.
     */
    get htmlElement() {
        const image = this._imageIndex;
        if (!image)
            return null;
        const wasm = this.engine.wasm;
        const jsImageIndex = wasm._wl_image_get_jsImage_index(image);
        /* Since the first element is `null`, no need to check for `jsImageIndex` */
        return wasm._images[jsImageIndex];
    }
    /**
     * Update a subrange on the texture to match the HTML element (e.g. reflect the current frame of a video).
     *
     * Usage:
     *
     * ```js
     * // Copies rectangle of pixel starting from (10, 20)
     * texture.updateSubImage(10, 20, 600, 400);
     * ```
     *
     * @param x x offset
     * @param y y offset
     * @param w width
     * @param h height
     */
    updateSubImage(x, y, w, h) {
        if (this.isDestroyed)
            return;
        const image = this._imageIndex;
        if (!image)
            return;
        const wasm = this.engine.wasm;
        const jsImageIndex = wasm._wl_image_get_jsImage_index(image);
        /* Lazy initialize temp canvas */
        if (!temp2d) {
            const canvas = document.createElement('canvas');
            const ctx = canvas.getContext('2d');
            if (!ctx) {
                throw new Error('Texture.updateSubImage(): Failed to obtain CanvasRenderingContext2D.');
            }
            temp2d = {
                canvas,
                ctx,
            };
        }
        const img = wasm._images[jsImageIndex];
        if (!img)
            return;
        /** @todo: If the image is an instance of canvas, this
         * can be skipped entirely. */
        temp2d.canvas.width = w;
        temp2d.canvas.height = h;
        temp2d.ctx.drawImage(img, x, y, w, h, 0, 0, w, h);
        const yOffset = (img.videoHeight ?? img.height) - y - h;
        wasm._images[jsImageIndex] = temp2d.canvas;
        wasm._wl_renderer_updateImage(image, x, yOffset);
        wasm._images[jsImageIndex] = img;
    }
    /**
     * Destroy and free the texture's texture altas space and memory.
     *
     * It is best practice to set the texture variable to `null` after calling
     * destroy to prevent accidental use of the invalid texture:
     *
     * ```js
     *   texture.destroy();
     *   texture = null;
     * ```
     *
     * @since 0.9.0
     */
    destroy() {
        const wasm = this.engine.wasm;
        wasm._wl_texture_destroy(this._id);
        this.engine.textures._destroy(this);
    }
    toString() {
        if (this.isDestroyed) {
            return 'Texture(destroyed)';
        }
        return `Texture(${this._index})`;
    }
    get _imageIndex() {
        return this.engine.wasm._wl_texture_get_image_index(this._id);
    }
}
/**
 * Wrapper around a native animation.
 */
export class Animation extends SceneResource {
    /**
     * @param index Index in the manager
     */
    constructor(host = WL, index) {
        const scene = host instanceof Prefab ? host : host.scene;
        super(scene, index);
    }
    /** Duration of this animation. */
    get duration() {
        return this.engine.wasm._wl_animation_get_duration(this._id);
    }
    /** Number of tracks in this animation. */
    get trackCount() {
        return this.engine.wasm._wl_animation_get_trackCount(this._id);
    }
    /**
     * Clone this animation retargeted to a new set of objects.
     *
     * The clone shares most of the data with the original and is therefore
     * light-weight.
     *
     * **Experimental:** This API might change in upcoming versions.
     *
     * If retargeting to {@link Skin}, the join names will be used to determine a mapping
     * from the previous skin to the new skin. The source skin will be retrieved from
     * the first track in the animation that targets a joint.
     *
     * @param newTargets New targets per track. Expected to have
     *      {@link Animation#trackCount} elements or to be a {@link Skin}.
     * @returns The retargeted clone of this animation.
     */
    retarget(newTargets) {
        const wasm = this.engine.wasm;
        if (newTargets instanceof Skin) {
            const index = wasm._wl_animation_retargetToSkin(this._id, newTargets._id);
            return this._scene.animations.wrap(index);
        }
        if (newTargets.length != this.trackCount) {
            throw Error('Expected ' +
                this.trackCount.toString() +
                ' targets, but got ' +
                newTargets.length.toString());
        }
        const ptr = wasm._malloc(2 * newTargets.length);
        for (let i = 0; i < newTargets.length; ++i) {
            const object3d = newTargets[i];
            this.scene.assertOrigin(object3d);
            /* Use local id here and **not** packed id, because the WASM
             * takes ownership and reinterprets the ids. */
            wasm.HEAPU16[ptr >> (1 + i)] = newTargets[i].objectId;
        }
        const index = wasm._wl_animation_retarget(this._id, ptr);
        wasm._free(ptr);
        return this._scene.animations.wrap(index);
    }
    toString() {
        if (this.isDestroyed) {
            return 'Animation(destroyed)';
        }
        return `Animation(${this._index})`;
    }
}
/**
 * Scene graph object.
 *
 * Node in the scene graph or "entity". Consists of transformation and a reference
 * to its parent object. Usually holds components and is accessible by components
 * through {@link Component#object}.
 *
 * Objects are stored in a data oriented manner inside WebAssembly memory. This class
 * is a JavaScript API wrapper around this memory for more convenient use in
 * components.
 *
 * Objects can be created and added to a scene through {@link Prefab.addObject}:
 *
 * ```js
 * const parent = scene.addObject();
 * parent.name = 'parent`';
 * const child = scene.addObject(parent);
 * child.name = 'child`';
 * ```
 */
export class Object3D {
    /**
     * Packed object id, containing scene index and local id.
     *
     * @hidden
     */
    _id = -1;
    /** Object id, relative to the scene manager. @hidden */
    _localId = -1;
    /** Scene instance. @hidden */
    _scene;
    /** Wonderland Engine instance. @hidden */
    _engine;
    /**
     * @param o Object id to wrap.
     *
     * @deprecated Objects must be obtained via {@link Scene.addObject} or {@link Scene.wrap}:
     *
     * ```js
     * // Create a new object.
     * const obj = scene.addObject();
     *
     * // Wrap an object using its id. The id must be valid.
     * const obj = scene.wrap(0);
     * ```
     *
     * @hidden
     */
    constructor(scene, id) {
        scene = scene instanceof Prefab ? scene : scene.scene;
        this._localId = id;
        this._id = (scene._index << 22) | id;
        this._scene = scene;
        /** @todo: Remove and only keep getter. */
        this._engine = scene.engine;
    }
    /**
     * Name of the object.
     *
     * Useful for identifying objects during debugging.
     */
    get name() {
        const wasm = this._engine.wasm;
        return wasm.UTF8ToString(wasm._wl_object_name(this._id));
    }
    /**
     * Set the object's name.
     *
     * @param newName The new name to set.
     */
    set name(newName) {
        const wasm = this._engine.wasm;
        wasm._wl_object_set_name(this._id, wasm.tempUTF8(newName));
    }
    /**
     * Parent of this object or `null` if parented to root.
     */
    get parent() {
        const p = this._engine.wasm._wl_object_parent(this._id);
        return p === 0 ? null : this._scene.wrap(p);
    }
    /**
     * Equivalent to {@link Object3D.getChildren}.
     *
     * @note Prefer to use {@link Object3D.getChildren} for performance.
     */
    get children() {
        return this.getChildren();
    }
    /** The number of children of this object. */
    get childrenCount() {
        return this._engine.wasm._wl_object_get_children_count(this._id);
    }
    /**
     * Reparent object to given object.
     *
     * @note Reparenting is not trivial and might have a noticeable performance impact.
     *
     * @param newParent New parent or `null` to parent to root
     */
    set parent(newParent) {
        this.scene.assertOrigin(newParent);
        this._engine.wasm._wl_object_set_parent(this._id, newParent == null ? 0 : newParent._id);
    }
    /** Local object id in the scene manager. */
    get objectId() {
        return this._localId;
    }
    /** Scene instance. */
    get scene() {
        return this._scene;
    }
    /** Hosting engine instance. */
    get engine() {
        return this._engine;
    }
    /**
     * Clone this hierarchy into a new one.
     *
     * Cloning copies the hierarchy structure, object names,
     * as well as components.
     *
     * JavaScript components are cloned using {@link Component.copy}. You can
     * override this method in your components.
     *
     * @param parent The parent for the cloned hierarchy or `null` to clone
     *     into the scene root. Defaults to `null`.
     *
     * @returns The clone of this object.
     */
    clone(parent = null) {
        this.scene.assertOrigin(parent);
        const engine = this._engine;
        const id = engine.wasm._wl_object_clone(this._id, parent ? parent._id : 0);
        return this._scene.wrap(id);
    }
    /**
     * Children of this object.
     *
     * @note Child order is **undefined**. No assumptions should be made
     * about the index of a specific object.
     *
     * If you need to access a specific child of this object, you can
     * use {@link Object3D.findByName}.
     *
     * When the object exists in the scene at editor time, prefer passing it as
     * a component property.
     *
     * @note When providing an output array, only `this.childrenCount` elements will be written.
     * The rest of the array will not be modified by this method.
     *
     * @param out Destination array, expected to have at least `this.childrenCount` elements.
     * @returns The `out` parameter.
     */
    getChildren(out = new Array(this.childrenCount)) {
        const childrenCount = this.childrenCount;
        if (childrenCount === 0)
            return out;
        const wasm = this._engine.wasm;
        wasm.requireTempMem(childrenCount * 2);
        this._engine.wasm._wl_object_get_children(this._id, wasm._tempMem, wasm._tempMemSize >> 1);
        for (let i = 0; i < childrenCount; ++i) {
            out[i] = this._scene.wrap(wasm._tempMemUint16[i]);
        }
        return out;
    }
    /**
     * Reset local transformation (translation, rotation and scaling) to identity.
     *
     * @returns Reference to self (for method chaining).
     */
    resetTransform() {
        this._engine.wasm._wl_object_reset_translation_rotation(this._id);
        this._engine.wasm._wl_object_reset_scaling(this._id);
        return this;
    }
    /**
     * Reset local position and rotation to identity.
     *
     * @returns Reference to self (for method chaining).
     */
    resetPositionRotation() {
        this._engine.wasm._wl_object_reset_translation_rotation(this._id);
        return this;
    }
    /** @deprecated Please use {@link Object3D.resetPositionRotation} instead. */
    resetTranslationRotation() {
        return this.resetPositionRotation();
    }
    /**
     * Reset local rotation, keep translation.
     *
     * @note To reset both rotation and translation, prefer
     *       {@link resetTranslationRotation}.
     *
     * @returns Reference to self (for method chaining).
     */
    resetRotation() {
        this._engine.wasm._wl_object_reset_rotation(this._id);
        return this;
    }
    /**
     * Reset local translation, keep rotation.
     *
     * @note To reset both rotation and translation, prefer
     *       {@link resetTranslationRotation}.
     *
     * @returns Reference to self (for method chaining).
     */
    resetPosition() {
        this._engine.wasm._wl_object_reset_translation(this._id);
        return this;
    }
    /** @deprecated Please use {@link Object3D.resetPosition} instead. */
    resetTranslation() {
        return this.resetPosition();
    }
    /**
     * Reset local scaling to identity (``[1.0, 1.0, 1.0]``).
     *
     * @returns Reference to self (for method chaining).
     */
    resetScaling() {
        this._engine.wasm._wl_object_reset_scaling(this._id);
        return this;
    }
    /** @deprecated Please use {@link Object3D.translateLocal} instead. */
    translate(v) {
        return this.translateLocal(v);
    }
    /**
     * Translate object by a vector in the parent's space.
     *
     * @param v Vector to translate by.
     *
     * @returns Reference to self (for method chaining).
     */
    translateLocal(v) {
        this._engine.wasm._wl_object_translate(this._id, v[0], v[1], v[2]);
        return this;
    }
    /**
     * Translate object by a vector in object space.
     *
     * @param v Vector to translate by.
     *
     * @returns Reference to self (for method chaining).
     */
    translateObject(v) {
        this._engine.wasm._wl_object_translate_obj(this._id, v[0], v[1], v[2]);
        return this;
    }
    /**
     * Translate object by a vector in world space.
     *
     * @param v Vector to translate by.
     *
     * @returns Reference to self (for method chaining).
     */
    translateWorld(v) {
        this._engine.wasm._wl_object_translate_world(this._id, v[0], v[1], v[2]);
        return this;
    }
    /** @deprecated Please use {@link Object3D.rotateAxisAngleDegLocal} instead. */
    rotateAxisAngleDeg(a, d) {
        this.rotateAxisAngleDegLocal(a, d);
        return this;
    }
    /**
     * Rotate around given axis by given angle (degrees) in local space.
     *
     * @param a Vector representing the rotation axis.
     * @param d Angle in degrees.
     *
     * @note If the object is translated the rotation will be around
     *     the parent. To rotate around the object origin, use
     *     {@link rotateAxisAngleDegObject}
     *
     * @see {@link rotateAxisAngleRad}
     *
     * @returns Reference to self (for method chaining).
     */
    rotateAxisAngleDegLocal(a, d) {
        this._engine.wasm._wl_object_rotate_axis_angle(this._id, a[0], a[1], a[2], d);
        return this;
    }
    /** @deprecated Please use {@link Object3D.rotateAxisAngleRadLocal} instead. */
    rotateAxisAngleRad(a, d) {
        return this.rotateAxisAngleRadLocal(a, d);
    }
    /**
     * Rotate around given axis by given angle (radians) in local space.
     *
     * @param a Vector representing the rotation axis.
     * @param d Angle in radians.
     *
     * @note If the object is translated the rotation will be around
     *     the parent. To rotate around the object origin, use
     *     {@link rotateAxisAngleDegObject}
     *
     * @see {@link rotateAxisAngleDeg}
     *
     * @returns Reference to self (for method chaining).
     */
    rotateAxisAngleRadLocal(a, d) {
        this._engine.wasm._wl_object_rotate_axis_angle_rad(this._id, a[0], a[1], a[2], d);
        return this;
    }
    /**
     * Rotate around given axis by given angle (degrees) in object space.
     *
     * @param a Vector representing the rotation axis.
     * @param d Angle in degrees.
     *
     * Equivalent to prepending a rotation quaternion to the object's
     * local transformation.
     *
     * @see {@link rotateAxisAngleRadObject}
     *
     * @returns Reference to self (for method chaining).
     */
    rotateAxisAngleDegObject(a, d) {
        this._engine.wasm._wl_object_rotate_axis_angle_obj(this._id, a[0], a[1], a[2], d);
        return this;
    }
    /**
     * Rotate around given axis by given angle (radians) in object space
     * Equivalent to prepending a rotation quaternion to the object's
     * local transformation.
     *
     * @param a Vector representing the rotation axis
     * @param d Angle in degrees
     *
     * @see {@link rotateAxisAngleDegObject}
     *
     * @returns Reference to self (for method chaining).
     */
    rotateAxisAngleRadObject(a, d) {
        this._engine.wasm._wl_object_rotate_axis_angle_rad_obj(this._id, a[0], a[1], a[2], d);
        return this;
    }
    /** @deprecated Please use {@link Object3D.rotateLocal} instead. */
    rotate(q) {
        this.rotateLocal(q);
        return this;
    }
    /**
     * Rotate by a quaternion.
     *
     * @param q the Quaternion to rotate by.
     *
     * @returns Reference to self (for method chaining).
     */
    rotateLocal(q) {
        this._engine.wasm._wl_object_rotate_quat(this._id, q[0], q[1], q[2], q[3]);
        return this;
    }
    /**
     * Rotate by a quaternion in object space.
     *
     * Equivalent to prepending a rotation quaternion to the object's
     * local transformation.
     *
     * @param q the Quaternion to rotate by.
     *
     * @returns Reference to self (for method chaining).
     */
    rotateObject(q) {
        this._engine.wasm._wl_object_rotate_quat_obj(this._id, q[0], q[1], q[2], q[3]);
        return this;
    }
    /** @deprecated Please use {@link Object3D.scaleLocal} instead. */
    scale(v) {
        this.scaleLocal(v);
        return this;
    }
    /**
     * Scale object by a vector in object space.
     *
     * @param v Vector to scale by.
     *
     * @returns Reference to self (for method chaining).
     */
    scaleLocal(v) {
        this._engine.wasm._wl_object_scale(this._id, v[0], v[1], v[2]);
        return this;
    }
    getPositionLocal(out = new Float32Array(3)) {
        const wasm = this._engine.wasm;
        /* Translation is different than rotation & scaling.
         * We can't simply read the memory. */
        wasm._wl_object_get_translation_local(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    getTranslationLocal(out = new Float32Array(3)) {
        return this.getPositionLocal(out);
    }
    getPositionWorld(out = new Float32Array(3)) {
        const wasm = this._engine.wasm;
        /* Translation is different than rotation & scaling.
         * We can't simply read the memory. */
        wasm._wl_object_get_translation_world(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    getTranslationWorld(out = new Float32Array(3)) {
        return this.getPositionWorld(out);
    }
    /**
     * Set local / object space position.
     *
     * Concatenates a new translation dual quaternion onto the existing rotation.
     *
     * @param v New local position array/vector, expected to have at least 3 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setPositionLocal(v) {
        this._engine.wasm._wl_object_set_translation_local(this._id, v[0], v[1], v[2]);
        return this;
    }
    /** @deprecated Please use {@link Object3D.setPositionLocal} instead. */
    setTranslationLocal(v) {
        return this.setPositionLocal(v);
    }
    /**
     * Set world space position.
     *
     * Applies the inverse parent transform with a new translation dual quaternion
     * which is concatenated onto the existing rotation.
     *
     * @param v New world position array/vector, expected to have at least 3 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setPositionWorld(v) {
        this._engine.wasm._wl_object_set_translation_world(this._id, v[0], v[1], v[2]);
        return this;
    }
    /** @deprecated Please use {@link Object3D.setPositionWorld} instead. */
    setTranslationWorld(v) {
        return this.setPositionWorld(v);
    }
    getScalingLocal(out = new Float32Array(3)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_scaling_local(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        return out;
    }
    /**
     * Set local / object space scaling.
     *
     * @param v New local scaling array/vector, expected to have at least 3 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setScalingLocal(v) {
        this._engine.wasm._wl_object_set_scaling_local(this._id, v[0], v[1], v[2]);
        return this;
    }
    getScalingWorld(out = new Float32Array(3)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_scaling_world(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        return out;
    }
    /**
     * Set World space scaling.
     *
     * @param v New world scaling array/vector, expected to have at least 3 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setScalingWorld(v) {
        this._engine.wasm._wl_object_set_scaling_world(this._id, v[0], v[1], v[2]);
        return this;
    }
    getRotationLocal(out = new Float32Array(4)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_local(this._id) / 4; /* Align F32 */
        /* The first 4 floats represent the rotation quaternion. */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        out[3] = wasm.HEAPF32[ptr + 3];
        return out;
    }
    /**
     * Set local space rotation.
     *
     * @param v New world rotation array/vector, expected to have at least 4 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setRotationLocal(v) {
        this._engine.wasm._wl_object_set_rotation_local(this._id, v[0], v[1], v[2], v[3]);
        return this;
    }
    getRotationWorld(out = new Float32Array(4)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_world(this._id) / 4; /* Align F32 */
        /* The first 4 floats represent the rotation quaternion. */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        out[3] = wasm.HEAPF32[ptr + 3];
        return out;
    }
    /**
     * Set local space rotation.
     *
     * @param v New world rotation array/vector, expected to have at least 4 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setRotationWorld(v) {
        this._engine.wasm._wl_object_set_rotation_world(this._id, v[0], v[1], v[2], v[3]);
        return this;
    }
    getTransformLocal(out = new Float32Array(8)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_local(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        out[3] = wasm.HEAPF32[ptr + 3];
        out[4] = wasm.HEAPF32[ptr + 4];
        out[5] = wasm.HEAPF32[ptr + 5];
        out[6] = wasm.HEAPF32[ptr + 6];
        out[7] = wasm.HEAPF32[ptr + 7];
        return out;
    }
    /**
     * Set local space rotation.
     *
     * @param v New local transform array, expected to have at least 8 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setTransformLocal(v) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_local(this._id) / 4; /* Align F32 */
        wasm.HEAPF32[ptr] = v[0];
        wasm.HEAPF32[ptr + 1] = v[1];
        wasm.HEAPF32[ptr + 2] = v[2];
        wasm.HEAPF32[ptr + 3] = v[3];
        wasm.HEAPF32[ptr + 4] = v[4];
        wasm.HEAPF32[ptr + 5] = v[5];
        wasm.HEAPF32[ptr + 6] = v[6];
        wasm.HEAPF32[ptr + 7] = v[7];
        this.setDirty();
        return this;
    }
    getTransformWorld(out = new Float32Array(8)) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_world(this._id) / 4; /* Align F32 */
        out[0] = wasm.HEAPF32[ptr];
        out[1] = wasm.HEAPF32[ptr + 1];
        out[2] = wasm.HEAPF32[ptr + 2];
        out[3] = wasm.HEAPF32[ptr + 3];
        out[4] = wasm.HEAPF32[ptr + 4];
        out[5] = wasm.HEAPF32[ptr + 5];
        out[6] = wasm.HEAPF32[ptr + 6];
        out[7] = wasm.HEAPF32[ptr + 7];
        return out;
    }
    /**
     * Set world space rotation.
     *
     * @param v New world transform array, expected to have at least 8 elements.
     *
     * @returns Reference to self (for method chaining).
     */
    setTransformWorld(v) {
        const wasm = this._engine.wasm;
        const ptr = wasm._wl_object_trans_world(this._id) / 4; /* Align F32 */
        wasm.HEAPF32[ptr] = v[0];
        wasm.HEAPF32[ptr + 1] = v[1];
        wasm.HEAPF32[ptr + 2] = v[2];
        wasm.HEAPF32[ptr + 3] = v[3];
        wasm.HEAPF32[ptr + 4] = v[4];
        wasm.HEAPF32[ptr + 5] = v[5];
        wasm.HEAPF32[ptr + 6] = v[6];
        wasm.HEAPF32[ptr + 7] = v[7];
        this._engine.wasm._wl_object_trans_world_to_local(this._id);
        return this;
    }
    /**
     * Local space transformation.
     *
     * @deprecated Please use {@link Object3D.setTransformLocal} and
     * {@link Object3D.getTransformLocal} instead.
     */
    get transformLocal() {
        const wasm = this._engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_object_trans_local(this._id), 8);
    }
    /**
     * Set local transform.
     *
     * @param t Local space transformation.
     *
     * @since 0.8.5
     *
     * @deprecated Please use {@link Object3D.setTransformLocal} and
     * {@link Object3D.getTransformLocal} instead.
     */
    set transformLocal(t) {
        this.transformLocal.set(t);
        this.setDirty();
    }
    /**
     * Global / world space transformation.
     *
     * May recompute transformations of the hierarchy of this object,
     * if they were changed by JavaScript components this frame.
     *
     * @deprecated Please use {@link Object3D.setTransformWorld} and
     * {@link Object3D.getTransformWorld} instead.
     */
    get transformWorld() {
        const wasm = this._engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_object_trans_world(this._id), 8);
    }
    /**
     * Set world transform.
     *
     * @param t Global / world space transformation.
     *
     * @since 0.8.5
     *
     * @deprecated Please use {@link Object3D.setTransformWorld} and
     * {@link Object3D.getTransformWorld} instead.
     */
    set transformWorld(t) {
        this.transformWorld.set(t);
        this._engine.wasm._wl_object_trans_world_to_local(this._id);
    }
    /**
     * Local / object space scaling.
     *
     * @deprecated Please use {@link Object3D.setScalingLocal} and
     * {@link Object3D.getScalingLocal} instead.
     */
    get scalingLocal() {
        const wasm = this._engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_object_scaling_local(this._id), 3);
    }
    /**
     * Set local space scaling.
     *
     * @param s Local space scaling.
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.setScalingLocal} and
     * {@link Object3D.getScalingLocal} instead.
     */
    set scalingLocal(s) {
        this.scalingLocal.set(s);
        this.setDirty();
    }
    /**
     * Global / world space scaling.
     *
     * May recompute transformations of the hierarchy of this object,
     * if they were changed by JavaScript components this frame.
     *
     * @deprecated Please use {@link Object3D.setScalingWorld} and
     * {@link Object3D.getScalingWorld} instead.
     */
    get scalingWorld() {
        const wasm = this._engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_object_scaling_world(this._id), 3);
    }
    /**
     * Set world space scaling.
     *
     * @param t World space scaling.
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.setScalingWorld} and
     * {@link Object3D.getScalingWorld} instead.
     */
    set scalingWorld(s) {
        this.scalingWorld.set(s);
        this._engine.wasm._wl_object_scaling_world_to_local(this._id);
    }
    /**
     * Local space rotation.
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.getRotationLocal} and
     * {@link Object3D.setRotationLocal} instead.
     */
    get rotationLocal() {
        return this.transformLocal.subarray(0, 4);
    }
    /**
     * Global / world space rotation
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.getRotationWorld} and
     * {@link Object3D.setRotationWorld} instead.
     */
    get rotationWorld() {
        return this.transformWorld.subarray(0, 4);
    }
    /**
     * Set local space rotation.
     *
     * @param r Local space rotation
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.getRotationLocal} and
     * {@link Object3D.setRotationLocal} instead.
     */
    set rotationLocal(r) {
        this._engine.wasm._wl_object_set_rotation_local(this._id, r[0], r[1], r[2], r[3]);
    }
    /**
     * Set world space rotation.
     *
     * @param r Global / world space rotation.
     *
     * @since 0.8.7
     *
     * @deprecated Please use {@link Object3D.getRotationWorld} and
     * {@link Object3D.setRotationWorld} instead.
     */
    set rotationWorld(r) {
        this._engine.wasm._wl_object_set_rotation_world(this._id, r[0], r[1], r[2], r[3]);
    }
    /** @deprecated Please use {@link Object3D.getForwardWorld} instead. */
    getForward(out) {
        return this.getForwardWorld(out);
    }
    /**
     * Compute the object's forward facing world space vector.
     *
     * The forward vector in object space is along the negative z-axis, i.e.,
     * `[0, 0, -1]`.
     *
     * @param out Destination array/vector, expected to have at least 3 elements.
     * @return The `out` parameter.
     */
    getForwardWorld(out) {
        out[0] = 0;
        out[1] = 0;
        out[2] = -1;
        this.transformVectorWorld(out);
        return out;
    }
    /** @deprecated Please use {@link Object3D.getUpWorld} instead. */
    getUp(out) {
        return this.getUpWorld(out);
    }
    /**
     * Compute the object's up facing world space vector.
     *
     * @param out Destination array/vector, expected to have at least 3 elements.
     * @return The `out` parameter.
     */
    getUpWorld(out) {
        out[0] = 0;
        out[1] = 1;
        out[2] = 0;
        this.transformVectorWorld(out);
        return out;
    }
    /** @deprecated Please use {@link Object3D.getRightWorld} instead. */
    getRight(out) {
        return this.getRightWorld(out);
    }
    /**
     * Compute the object's right facing world space vector.
     *
     * @param out Destination array/vector, expected to have at least 3 elements.
     * @return The `out` parameter.
     */
    getRightWorld(out) {
        out[0] = 1;
        out[1] = 0;
        out[2] = 0;
        this.transformVectorWorld(out);
        return out;
    }
    /**
     * Transform a vector by this object's world transform.
     *
     * @param out Out vector
     * @param v Vector to transform, default `out`
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformVectorWorld(out, v = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = v[0];
        wasm._tempMemFloat[1] = v[1];
        wasm._tempMemFloat[2] = v[2];
        wasm._wl_object_transformVectorWorld(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a vector by this object's local transform.
     *
     * @param out Out vector
     * @param v Vector to transform, default `out`
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformVectorLocal(out, v = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = v[0];
        wasm._tempMemFloat[1] = v[1];
        wasm._tempMemFloat[2] = v[2];
        wasm._wl_object_transformVectorLocal(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a point by this object's world transform.
     *
     * @param out Out point.
     * @param p Point to transform, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformPointWorld(out, p = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = p[0];
        wasm._tempMemFloat[1] = p[1];
        wasm._tempMemFloat[2] = p[2];
        wasm._wl_object_transformPointWorld(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a point by this object's local transform.
     *
     * @param out Out point.
     * @param p Point to transform, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformPointLocal(out, p = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = p[0];
        wasm._tempMemFloat[1] = p[1];
        wasm._tempMemFloat[2] = p[2];
        wasm._wl_object_transformPointLocal(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a vector by this object's inverse world transform.
     *
     * @param out Out vector.
     * @param v Vector to transform, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformVectorInverseWorld(out, v = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = v[0];
        wasm._tempMemFloat[1] = v[1];
        wasm._tempMemFloat[2] = v[2];
        wasm._wl_object_transformVectorInverseWorld(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a vector by this object's inverse local transform.
     *
     * @param out Out vector
     * @param v Vector to transform, default `out`
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformVectorInverseLocal(out, v = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = v[0];
        wasm._tempMemFloat[1] = v[1];
        wasm._tempMemFloat[2] = v[2];
        wasm._wl_object_transformVectorInverseLocal(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a point by this object's inverse world transform.
     *
     * @param out Out point.
     * @param p Point to transform, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformPointInverseWorld(out, p = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat[0] = p[0];
        wasm._tempMemFloat[1] = p[1];
        wasm._tempMemFloat[2] = p[2];
        wasm._wl_object_transformPointInverseWorld(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform a point by this object's inverse local transform.
     *
     * @param out Out point.
     * @param p Point to transform, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    transformPointInverseLocal(out, p = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat.set(p);
        wasm._wl_object_transformPointInverseLocal(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        return out;
    }
    /**
     * Transform an object space dual quaternion into world space.
     *
     * @param out Out transformation.
     * @param q Local space transformation, default `out`.
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    toWorldSpaceTransform(out, q = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat.set(q);
        wasm._wl_object_toWorldSpaceTransform(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        out[3] = wasm._tempMemFloat[3];
        out[4] = wasm._tempMemFloat[4];
        out[5] = wasm._tempMemFloat[5];
        out[6] = wasm._tempMemFloat[6];
        out[7] = wasm._tempMemFloat[7];
        return out;
    }
    /**
     * Transform a world space dual quaternion into local space.
     *
     * @param out Out transformation
     * @param q World space transformation, default `out`
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    toLocalSpaceTransform(out, q = out) {
        const p = this.parent;
        if (p) {
            p.toObjectSpaceTransform(out, q);
            return out;
        }
        if (out !== q) {
            out[0] = q[0];
            out[1] = q[1];
            out[2] = q[2];
            out[3] = q[3];
            out[4] = q[4];
            out[5] = q[5];
            out[6] = q[6];
            out[7] = q[7];
        }
        return out;
    }
    /**
     * Transform a world space dual quaternion into object space.
     *
     * @param out Out transformation.
     * @param q World space transformation, default `out`
     * @return The `out` parameter.
     *
     * @since 0.8.7
     */
    toObjectSpaceTransform(out, q = out) {
        const wasm = this._engine.wasm;
        wasm._tempMemFloat.set(q);
        wasm._wl_object_toObjectSpaceTransform(this._id, wasm._tempMem);
        out[0] = wasm._tempMemFloat[0];
        out[1] = wasm._tempMemFloat[1];
        out[2] = wasm._tempMemFloat[2];
        out[3] = wasm._tempMemFloat[3];
        out[4] = wasm._tempMemFloat[4];
        out[5] = wasm._tempMemFloat[5];
        out[6] = wasm._tempMemFloat[6];
        out[7] = wasm._tempMemFloat[7];
        return out;
    }
    /**
     * Turn towards / look at target.
     *
     * Rotates the object so that its forward vector faces towards the target
     * position. The `up` vector acts as a hint to uniquely orient the object's
     * up direction. When orienting a view component, the projected `up` vector
     * faces upwards on the viewing plane.
     *
     * @param p Target position to turn towards, in world space.
     * @param up Up vector to align object with, in world space. Default is `[0, 1, 0]`.
     *
     * @returns Reference to self (for method chaining).
     */
    lookAt(p, up = UP_VECTOR) {
        this._engine.wasm._wl_object_lookAt(this._id, p[0], p[1], p[2], up[0], up[1], up[2]);
        return this;
    }
    /** Destroy the object with all of its components and remove it from the scene */
    destroy() {
        if (this._id < 0)
            return;
        /* This will automatically call `scene._destroyObject` */
        this.engine.wasm._wl_object_remove(this._id);
    }
    /**
     * Mark transformation dirty.
     *
     * Causes an eventual recalculation of {@link transformWorld}, either
     * on next {@link getTranslationWorld}, {@link transformWorld} or
     * {@link scalingWorld} or the beginning of next frame, whichever
     * happens first.
     */
    setDirty() {
        this._engine.wasm._wl_object_set_dirty(this._id);
    }
    /**
     * Disable/enable all components of this object.
     *
     * @param b New state for the components.
     *
     * @since 0.8.5
     */
    set active(b) {
        const comps = this.getComponents();
        for (let c of comps) {
            c.active = b;
        }
    }
    getComponent(typeOrClass, index = 0) {
        const wasm = this._engine.wasm;
        const type = isString(typeOrClass) ? typeOrClass : typeOrClass.TypeName;
        const scene = this._scene;
        const componentType = wasm._wl_scene_get_component_manager_index(scene._index, wasm.tempUTF8(type));
        if (componentType < 0) {
            /* Not a native component, try js: */
            const typeIndex = wasm._componentTypeIndices[type];
            if (typeIndex === undefined)
                return null;
            const jsIndex = wasm._wl_get_js_component_index(this._id, typeIndex, index);
            if (jsIndex < 0)
                return null;
            const component = this._scene._jsComponents[jsIndex];
            return component.constructor !== BrokenComponent ? component : null;
        }
        const componentId = wasm._wl_get_component_id(this._id, componentType, index);
        return scene._components.wrapNative(componentType, componentId);
    }
    getComponents(typeOrClass) {
        const wasm = this._engine.wasm;
        const scene = this._scene;
        let manager = null;
        let type = null;
        if (typeOrClass) {
            type = isString(typeOrClass) ? typeOrClass : typeOrClass.TypeName;
            const nativeManager = scene._components.getNativeManager(type);
            manager = nativeManager !== null ? nativeManager : scene._components.js;
        }
        const components = [];
        const maxComps = Math.floor((wasm._tempMemSize / 3) * 2);
        const componentsCount = wasm._wl_object_get_components(this._id, wasm._tempMem, maxComps);
        const offset = 2 * componentsCount;
        wasm._wl_object_get_component_types(this._id, wasm._tempMem + offset, maxComps);
        for (let i = 0; i < componentsCount; ++i) {
            const t = wasm._tempMemUint8[i + offset];
            const componentId = wasm._tempMemUint16[i];
            if (manager !== null && t !== manager)
                continue;
            const comp = this._scene._components.wrapAny(t, componentId);
            if (!comp)
                continue;
            if (type && type !== comp.constructor.TypeName)
                continue;
            components.push(comp);
        }
        return components;
    }
    addComponent(typeOrClass, params) {
        const wasm = this._engine.wasm;
        const type = isString(typeOrClass) ? typeOrClass : typeOrClass.TypeName;
        const nativeManager = this._scene._components.getNativeManager(type);
        const isNative = nativeManager !== null;
        const manager = isNative ? nativeManager : this._scene._components.js;
        let componentId = -1;
        if (!isNative) {
            /* JavaScript component */
            if (!(type in wasm._componentTypeIndices)) {
                throw new TypeError("Unknown component type '" + type + "'");
            }
            componentId = wasm._wl_object_add_js_component(this._id, wasm._componentTypeIndices[type]);
        }
        else {
            /* native component */
            componentId = wasm._wl_object_add_component(this._id, manager);
        }
        const component = this._scene._components.wrapAny(manager, componentId);
        if (params !== undefined)
            component.copy(params);
        /* Explicitly initialize js components */
        if (!isNative) {
            component._triggerInit();
            /* start() is called through onActivate() */
        }
        /* If it was not explicitly requested by the user to leave the component inactive,
         * we activate it as a final step. This invalidates componentIndex! */
        if (!params || !('active' in params && !params.active)) {
            component.active = true;
        }
        return component;
    }
    /**
     * Search for descendants matching the name.
     *
     * This method is a wrapper around {@link Object3D.findByNameDirect} and
     * {@link Object3D.findByNameRecursive}.
     *
     * @param name The name to search for.
     * @param recursive If `true`, the method will look at all the descendants of this object.
     *     If `false`, this method will only perform the search in direct children.
     * @returns An array of {@link Object3D} matching the name.
     *
     * @since 1.1.0
     */
    findByName(name, recursive = false) {
        return recursive ? this.findByNameRecursive(name) : this.findByNameDirect(name);
    }
    /**
     * Search for all **direct** children matching the name.
     *
     * @note Even though this method is heavily optimized, it does perform
     * a linear search to find the objects. Do not use in a hot path.
     *
     * @param name The name to search for.
     * @returns An array of {@link Object3D} matching the name.
     *
     * @since 1.1.0
     */
    findByNameDirect(name) {
        const wasm = this._engine.wasm;
        const id = this._id;
        /* Divide by 4 to get half as many ushort as possible */
        const tempSizeU16 = wasm._tempMemSize >> 2;
        const maxCount = tempSizeU16 - 2; /* Reserve two ushort */
        const buffer = wasm._tempMemUint16;
        buffer[maxCount] = 0; /* Index offset */
        buffer[maxCount + 1] = 0; /* child count */
        const bufferPtr = wasm._tempMem;
        const indexPtr = bufferPtr + maxCount * 2;
        const childCountPtr = bufferPtr + maxCount * 2 + 2;
        const namePtr = wasm.tempUTF8(name, (maxCount + 2) * 2);
        const result = [];
        let read = 0;
        while ((read = wasm._wl_object_findByName(id, namePtr, indexPtr, childCountPtr, bufferPtr, maxCount))) {
            for (let i = 0; i < read; ++i) {
                result.push(this._scene.wrap(buffer[i]));
            }
        }
        return result;
    }
    /**
     * Search for **all descendants** matching the name.
     *
     * @note Even though this method is heavily optimized, it does perform
     * a linear search to find the objects. Do not use in a hot path.
     *
     * @param name The name to search for.
     * @returns An array of {@link Object3D} matching the name.
     *
     * @since 1.1.0
     */
    findByNameRecursive(name) {
        const wasm = this._engine.wasm;
        const id = this._id;
        /* Divide by 4 to get half as many ushort as possible */
        const tempSizeU16 = wasm._tempMemSize >> 2;
        const maxCount = tempSizeU16 - 1; /* Reserve one ushort */
        const buffer = wasm._tempMemUint16;
        buffer[maxCount] = 0; /* Index offset */
        const bufferPtr = wasm._tempMem;
        const indexPtr = bufferPtr + maxCount * 2;
        const namePtr = wasm.tempUTF8(name, (maxCount + 1) * 2);
        let read = 0;
        const result = [];
        while ((read = wasm._wl_object_findByNameRecursive(id, namePtr, indexPtr, bufferPtr, maxCount))) {
            for (let i = 0; i < read; ++i) {
                result.push(this._scene.wrap(buffer[i]));
            }
        }
        return result;
    }
    /**
     * Whether given object's transformation has changed.
     */
    get changed() {
        return !!this._engine.wasm._wl_object_is_changed(this._id);
    }
    /**
     * `true` if the object is destroyed, `false` otherwise.
     *
     * If {@link WonderlandEngine.erasePrototypeOnDestroy} is `true`,
     * reading a custom property will not work:
     *
     * ```js
     * engine.erasePrototypeOnDestroy = true;
     *
     * const obj = scene.addObject();
     * obj.customParam = 'Hello World!';
     *
     * console.log(obj.isDestroyed); // Prints `false`
     * obj.destroy();
     * console.log(obj.isDestroyed); // Prints `true`
     * console.log(obj.customParam); // Throws an error
     * ```
     *
     * @since 1.1.1
     */
    get isDestroyed() {
        return this._id < 0;
    }
    /**
     * Checks equality by comparing ids and **not** the JavaScript reference.
     *
     * @deprecate Use JavaScript reference comparison instead:
     *
     * ```js
     * const objectA = scene.addObject();
     * const objectB = scene.addObject();
     * const objectC = objectB;
     * console.log(objectA === objectB); // false
     * console.log(objectA === objectA); // true
     * console.log(objectB === objectC); // true
     * ```
     */
    equals(otherObject) {
        /** @todo(2.0.0): Remove this method. */
        if (!otherObject)
            return false;
        return this._id == otherObject._id;
    }
    toString() {
        if (this.isDestroyed) {
            return 'Object3D(destroyed)';
        }
        return `Object3D('${this.name}', ${this._localId})`;
    }
}
/**
 * Wrapper around a native skin data.
 */
export class Skin extends SceneResource {
    /** Amount of joints in this skin. */
    get jointCount() {
        return this.engine.wasm._wl_skin_get_joint_count(this._id);
    }
    /** Joints object ids for this skin */
    get jointIds() {
        const wasm = this.engine.wasm;
        return new Uint16Array(wasm.HEAPU16.buffer, wasm._wl_skin_joint_ids(this._id), this.jointCount);
    }
    /**
     * Dual quaternions in a flat array of size 8 times {@link jointCount}.
     *
     * Inverse bind transforms of the skin.
     */
    get inverseBindTransforms() {
        const wasm = this.engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_skin_inverse_bind_transforms(this._id), 8 * this.jointCount);
    }
    /**
     * Vectors in a flat array of size 3 times {@link jointCount}.
     *
     * Inverse bind scalings of the skin.
     */
    get inverseBindScalings() {
        const wasm = this.engine.wasm;
        return new Float32Array(wasm.HEAPF32.buffer, wasm._wl_skin_inverse_bind_scalings(this._id), 3 * this.jointCount);
    }
}
/**
 * Wrapper around a native set of morph targets.
 *
 * ## Usage
 *
 * ```js
 * const mesh = object.getComponent('mesh');
 * const mouthTarget = mesh.morphTargets.getTargetIndex('mouth');
 * mesh.setMorphTargetWeight(mouthTarget, 0.5);
 * ```
 *
 * @since 1.2.0
 */
export class MorphTargets extends Resource {
    /** Amount of targets in this morph target set. */
    get count() {
        return this.engine.wasm._wl_morph_targets_get_target_count(this._id);
    }
    /** Returns the name of a given target */
    getTargetName(target) {
        if (target >= this.count) {
            throw new Error(`Index ${target} is out of bounds for ${this.count} targets`);
        }
        const wasm = this.engine.wasm;
        return wasm.UTF8ToString(wasm._wl_morph_targets_get_target_name(this._id, target));
    }
    /**
     * Get the index for a given target name.
     *
     * Throws if no target with that name exists.
     *
     * @param name Name of the target.
     */
    getTargetIndex(name) {
        const wasm = this.engine.wasm;
        const index = wasm._wl_morph_targets_get_target_index(this._id, wasm.tempUTF8(name));
        if (index === -1) {
            throw Error(`Missing target '${name}'`);
        }
        return index;
    }
}
/* For backward compatibility with < 1.0.0. */
export { Object3D as Object };
/**
 * Ray hit.
 *
 * Result of a {@link Scene.rayCast} or {@link Physics.rayCast}.
 *
 * @note this class wraps internal engine data and should only be created internally.
 */
export class RayHit {
    /** Scene instance. @hidden */
    _scene;
    /** Pointer to the memory heap. */
    _ptr;
    /**
     * @param ptr Pointer to the ray hits memory.
     */
    constructor(scene, ptr) {
        if ((ptr & 3) !== 0) {
            throw new Error('Misaligned pointer: please report a bug');
        }
        this._scene = scene;
        this._ptr = ptr;
    }
    getLocations(out) {
        out = out ?? Array.from({ length: this.hitCount }, () => new Float32Array(3));
        const wasm = this.engine.wasm;
        const alignedPtr = this._ptr / 4; /* Align F32 */
        for (let i = 0; i < this.hitCount; ++i) {
            const locationPtr = alignedPtr + 3 * i;
            out[i][0] = wasm.HEAPF32[locationPtr];
            out[i][1] = wasm.HEAPF32[locationPtr + 1];
            out[i][2] = wasm.HEAPF32[locationPtr + 2];
        }
        return out;
    }
    getNormals(out) {
        out = out ?? Array.from({ length: this.hitCount }, () => new Float32Array(3));
        const wasm = this.engine.wasm;
        const alignedPtr = (this._ptr + 48) / 4; /* Align F32 */
        for (let i = 0; i < this.hitCount; ++i) {
            const normalPtr = alignedPtr + 3 * i;
            out[i][0] = wasm.HEAPF32[normalPtr];
            out[i][1] = wasm.HEAPF32[normalPtr + 1];
            out[i][2] = wasm.HEAPF32[normalPtr + 2];
        }
        return out;
    }
    getDistances(out = new Float32Array(this.hitCount)) {
        const wasm = this.engine.wasm;
        const alignedPtr = (this._ptr + 48 * 2) / 4; /* Align F32 */
        for (let i = 0; i < this.hitCount; ++i) {
            const distancePtr = alignedPtr + i;
            out[i] = wasm.HEAPF32[distancePtr];
        }
        return out;
    }
    /**
     * Array of hit objects.
     *
     * @param out Destination array/vector, expected to have at least `this.hitCount` elements.
     * @returns The `out` parameter.
     */
    getObjects(out = new Array(this.hitCount)) {
        const HEAPU16 = this.engine.wasm.HEAPU16;
        const alignedPtr = (this._ptr + (48 * 2 + 16)) >> 1;
        for (let i = 0; i < this.hitCount; ++i) {
            out[i] = this._scene.wrap(HEAPU16[alignedPtr + i]);
        }
        return out;
    }
    /** Hosting engine instance. */
    get engine() {
        return this._scene.engine;
    }
    /**
     * Equivalent to {@link RayHit.getLocations}.
     *
     * @note Prefer to use {@link RayHit.getLocations} for performance.
     */
    get locations() {
        return this.getLocations();
    }
    /**
     * Equivalent to {@link RayHit.getNormals}.
     *
     * @note Prefer to use {@link RayHit.getNormals} for performance.
     */
    get normals() {
        return this.getNormals();
    }
    /**
     * Equivalent to {@link RayHit.getDistances}.
     *
     * @note Prefer to use {@link RayHit.getDistances} for performance.
     */
    get distances() {
        return this.getDistances();
    }
    /**
     * Equivalent to {@link RayHit.getObjects}.
     *
     * @note Prefer to use {@link RayHit.getObjects} for performance.
     */
    get objects() {
        /** @todo: Remove at 2.0.0, this is kept for backward compatibility. */
        const objects = [null, null, null, null];
        return this.getObjects(objects);
    }
    /** Number of hits (max 4) */
    get hitCount() {
        return Math.min(this.engine.wasm.HEAPU32[this._ptr / 4 + 30], 4);
    }
}
class math {
    /** (Experimental!) Cubic Hermite spline interpolation for vector3 and quaternions.
     *
     * With `f == 0`, `out` will be `b`, if `f == 1`, `out` will be c.
     *
     * Whether a quaternion or vector3 interpolation is intended is determined by
     * length of `a`.
     *
     * @param out Array to write result to.
     * @param a First tangent/handle.
     * @param b First point or quaternion.
     * @param c Second point or quaternion.
     * @param d Second handle.
     * @param f Interpolation factor in [0; 1].
     * @returns The `out` parameter.
     *
     * @since 0.8.6
     */
    static cubicHermite(out, a, b, c, d, f, engine = WL) {
        const wasm = engine.wasm;
        wasm._tempMemFloat.subarray(0).set(a);
        wasm._tempMemFloat.subarray(4).set(b);
        wasm._tempMemFloat.subarray(8).set(c);
        wasm._tempMemFloat.subarray(12).set(d);
        const isQuat = a.length == 4;
        wasm._wl_math_cubicHermite(wasm._tempMem + 4 * 16, wasm._tempMem + 4 * 0, wasm._tempMem + 4 * 4, wasm._tempMem + 4 * 8, wasm._tempMem + 4 * 12, f, isQuat);
        out[0] = wasm._tempMemFloat[16];
        out[1] = wasm._tempMemFloat[17];
        out[2] = wasm._tempMemFloat[18];
        if (isQuat)
            out[3] = wasm._tempMemFloat[19];
        return out;
    }
}
export { math };
/**
 * Class for accessing internationalization (i18n) features.
 *
 * Allows {@link I18N.onLanguageChanged "detecting language change"},
 * {@link I18N.language "setting the current language"} or translating
 * {@link I18N.translate "individual terms"}.
 *
 * Internationalization works with terms,
 * a string type keyword that is linked to a different text for each language.
 *
 * Internally, string parameters for text and js components are
 * automatically swapped during language change, given they are linked to a term.
 * If manual text swapping is desired, {@link I18N.translate}
 * can be used to retrieve the current translation for any term.
 *
 * You can also use the {@link I18N.onLanguageChanged} to manually update text
 * when a language is changed to for example update a number in a string.
 *
 * @since 1.0.0
 */
export class I18N {
    /**
     * {@link Emitter} for language change events.
     *
     * First parameter to a listener is the old language index,
     * second parameter is the new language index.
     *
     * Usage from a within a component:
     *
     * ```js
     * this.engine.i18n.onLanguageChanged.add((oldLanguageIndex, newLanguageIndex) => {
     *     const oldLanguage = this.engine.i18n.languageName(oldLanguageIndex);
     *     const newLanguage = this.engine.i18n.languageName(newLanguageIndex);
     *     console.log("Switched from", oldLanguage, "to", newLanguage);
     * });
     * ```
     */
    onLanguageChanged = new Emitter();
    /** Wonderland Engine instance. @hidden */
    _engine;
    /** Previously set language index. @hidden */
    _prevLanguageIndex = -1;
    /**
     * Constructor
     */
    constructor(engine) {
        this._engine = engine;
    }
    /**
     * Set current language and apply translations to linked text parameters.
     *
     * @note This is equivalent to {@link I18N.setLanguage}.
     *
     * @param code Language code to switch to
     */
    set language(code) {
        this.setLanguage(code);
    }
    /** Get current language code. */
    get language() {
        const wasm = this._engine.wasm;
        const code = wasm._wl_i18n_currentLanguage();
        if (code === 0)
            return null;
        return wasm.UTF8ToString(code);
    }
    /**
     * Get the current language index.
     *
     * This method is more efficient than its equivalent:
     *
     * ```js
     * const index = i18n.languageIndex(i18n.language);
     * ```
     */
    get currentIndex() {
        return this._engine.wasm._wl_i18n_currentLanguageIndex();
    }
    /** Previous language index. */
    get previousIndex() {
        return this._prevLanguageIndex;
    }
    /**
     * Set current language and apply translations to linked text parameters.
     *
     * @param code The language code.
     * @returns A promise that resolves with the current index code when the
     *     language is loaded.
     */
    async setLanguage(code) {
        if (code == null)
            return Promise.resolve(this.currentIndex);
        const wasm = this._engine.wasm;
        this._prevLanguageIndex = this.currentIndex;
        wasm._wl_i18n_setLanguage(wasm.tempUTF8(code));
        const scene = this.engine.scene;
        const filename = wasm.UTF8ToString(wasm._wl_i18n_languageFile(this.currentIndex));
        const url = `${scene.baseURL}/locale/${filename}`;
        await scene._downloadDependency(url);
        this.onLanguageChanged.notify(this._prevLanguageIndex, this.currentIndex);
        return this.currentIndex;
    }
    /**
     * Get translated string for a term for the currently loaded language.
     *
     * @param term Term to translate
     */
    translate(term) {
        const wasm = this._engine.wasm;
        const translation = wasm._wl_i18n_translate(wasm.tempUTF8(term));
        if (translation === 0)
            return null;
        return wasm.UTF8ToString(translation);
    }
    /**
     * Get the number of languages in the project.
     *
     */
    languageCount() {
        const wasm = this._engine.wasm;
        return wasm._wl_i18n_languageCount();
    }
    /**
     * Get a language code.
     *
     * @param index Index of the language to get the code from
     */
    languageIndex(code) {
        const wasm = this._engine.wasm;
        return wasm._wl_i18n_languageIndex(wasm.tempUTF8(code));
    }
    /**
     * Get a language code.
     *
     * @param index Index of the language to get the code from
     */
    languageCode(index) {
        const wasm = this._engine.wasm;
        const code = wasm._wl_i18n_languageCode(index);
        if (code === 0)
            return null;
        return wasm.UTF8ToString(code);
    }
    /**
     * Get a language name.
     *
     * @param index Index of the language to get the name from
     */
    languageName(index) {
        const wasm = this._engine.wasm;
        const name = wasm._wl_i18n_languageName(index);
        if (name === 0)
            return null;
        return wasm.UTF8ToString(name);
    }
    /** Hosting engine instance. */
    get engine() {
        return this._engine;
    }
}
/** Properties of a WebXR session */
export class XR {
    /** Wonderland WASM bridge. @hidden */
    #wasm;
    #mode;
    constructor(wasm, mode) {
        this.#wasm = wasm;
        this.#mode = mode;
    }
    /** Current WebXR session mode */
    get sessionMode() {
        return this.#mode;
    }
    /** Current WebXR session */
    get session() {
        return this.#wasm.webxr_session;
    }
    /** Current WebXR frame */
    get frame() {
        return this.#wasm.webxr_frame;
    }
    referenceSpaceForType(type) {
        return this.#wasm.webxr_refSpaces[type] ?? null;
    }
    /** Set current reference space type used for retrieving eye, head, hand and joint poses */
    set currentReferenceSpace(refSpace) {
        this.#wasm.webxr_refSpace = refSpace;
        this.#wasm.webxr_refSpaceType = null;
        for (const type of Object.keys(this.#wasm.webxr_refSpaces)) {
            if (this.#wasm.webxr_refSpaces[type] === refSpace) {
                /* Keep track of reference space type */
                this.#wasm.webxr_refSpaceType = type;
            }
        }
    }
    /** Current reference space type used for retrieving eye, head, hand and joint poses */
    get currentReferenceSpace() {
        return this.#wasm.webxr_refSpace;
    }
    /** Current WebXR reference space type or `null` if not a default reference space */
    get currentReferenceSpaceType() {
        return this.#wasm.webxr_refSpaceType;
    }
    /** Current WebXR base layer  */
    get baseLayer() {
        return this.#wasm.webxr_baseLayer;
    }
    /** Current WebXR framebuffer */
    get framebuffers() {
        if (!Array.isArray(this.#wasm.webxr_fbo)) {
            return [this.#wasm.GL.framebuffers[this.#wasm.webxr_fbo]];
        }
        return this.#wasm.webxr_fbo.map((id) => this.#wasm.GL.framebuffers[id]);
    }
}
