/**
 * Downloads a video from URL and applies it as `diffuseTexture` or `flatTexture`
 * on given material.
 *
 * Video textures need to be updated regularly whenever
 * a new frame is available. This component handles the
 * detection of a new frame and updates the texture to
 * reflect the video's current frame.
 *
 * Materials from the following shaders are supported:
 *  - "Phong Opaque Textured"
 *  - "Flat Opaque Textured"
 *  - "Background"
 *  - "Physical Opaque Textured"
 *  - "Foliage"
 *
 * The video can be accessed through `this.video`:
 *
 * ```js
 *   let videoTexture = this.object.getComponent('video-texture');
 *   videoTexture.video.play();
 *   videoTexture.video.pause();
 * ```
 *
 * See [Video Example](/showcase/video).
 */
export class VideoTexture extends Component {
    static Properties: {
        /** URL to download video from */
        url: import("@wonderlandengine/api").ComponentProperty;
        /** Material to apply the video texture to */
        material: import("@wonderlandengine/api").ComponentProperty;
        /** Whether to loop the video */
        loop: import("@wonderlandengine/api").ComponentProperty;
        /** Whether to automatically start playing the video */
        autoplay: import("@wonderlandengine/api").ComponentProperty;
        /** Whether to mute sound */
        muted: import("@wonderlandengine/api").ComponentProperty;
        /** Name of the texture property to set */
        textureProperty: import("@wonderlandengine/api").ComponentProperty;
    };
    init(): void;
    loaded: boolean | undefined;
    frameUpdateRequested: boolean | undefined;
    start(): void;
    video: HTMLVideoElement | undefined;
    applyTexture(): void;
    texture: Texture | undefined;
    update(dt: any): void;
    updateVideo(): void;
}
import { Component } from '@wonderlandengine/api';
import { Texture } from '@wonderlandengine/api';
